#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/material/get_materials_by_formula.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Overview
# 
# Inside this example we contact [Material](https://docs.mat3ra.com/api/Material/get_materials) endpoint to obtain a list materials that an account has access to. We use chemical formula to filter the list.

# # Complete Authorization Form and Initialize Settings
# 
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
# 
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
# 
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
# 
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
# 
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
# 
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# # Imports

# In[ ]:


from utils.settings import ENDPOINT_ARGS, ACCOUNT_ID
from utils.generic import display_JSON

from exabyte_api_client.endpoints.materials import MaterialEndpoints


# ## Set Parameters
# 
# - **QUERY**: A query describing the documents to find. See [Meteor collection](https://docs.meteor.com/api/collections.html#Mongo-Collection-find) for more information.

# In[ ]:


QUERY = {"formula": "Si", "owner._id": ACCOUNT_ID}


# ## Initialize the endpoint

# In[ ]:


endpoint = MaterialEndpoints(*ENDPOINT_ARGS)


# ## List materials
# 
# Contact the endpoint to list materials according to the query above.

# In[ ]:


materials = endpoint.list(QUERY)


# ## Print materials
# 
# Print the list of materials saved under the corresponding variable in pretty JSON below.

# In[ ]:


display_JSON(materials)




import io
import os
import re

from notebook.utils import to_api_path

_script_exporter = None


def script_post_save(model, os_path, contents_manager, **kwargs):
    """
    Converts notebooks to Python script on save.
    See https://jupyter-notebook.readthedocs.io/en/stable/extending/savehooks.html for more information.
    """
    from nbconvert.exporters.script import ScriptExporter

    if model["type"] != "notebook":
        return

    global _script_exporter

    if _script_exporter is None:
        _script_exporter = ScriptExporter(parent=contents_manager)

    log = contents_manager.log

    base, ext = os.path.splitext(os.path.realpath(os_path))
    script, resources = _script_exporter.from_filename(os_path)
    script_fname = base + resources.get("output_extension", ".txt")
    log.info("Saving script /%s", to_api_path(script_fname, contents_manager.root_dir))

    # Remove notebook numberings from auto-generated .py files
    pattern = """
    (?<=^\#\sIn\[)   # Lookbehind to ensure we get the "# In[" bits notebooks put at the start of these lines
    (\d+)           # 1 or more digits, since we don't care about changing "[]" portions
    (?=\]:)         # Lookahead to get to the closing "]:" at the end of the line
    """
    script = re.sub(pattern, "", script, flags=re.MULTILINE | re.VERBOSE)

    with io.open(script_fname, "w", encoding="utf-8") as f:
        f.write(script)


c.FileContentsManager.post_save_hook = script_post_save  # noqa F821



import io
import numpy as np
from pymatgen.core import Structure, Lattice
from pymatgen.io.ase import AseAtomsAdaptor
from ase import Atoms
from ase.io import read, write
from ase.calculators.calculator import Calculator


def to_pymatgen(material_data):
    """
    Converts material object in ESSE format to a pymatgen Structure object.

    Args:
        material_data (dict): A dictionary containing the material information in ESSE format.

    Returns:
        Structure: A pymatgen Structure object.
    """

    # Extract lattice information
    lattice_params = material_data["lattice"]
    a = lattice_params["a"]
    b = lattice_params["b"]
    c = lattice_params["c"]
    alpha = lattice_params["alpha"]
    beta = lattice_params["beta"]
    gamma = lattice_params["gamma"]

    # Create a Lattice from parameters
    lattice = Lattice.from_parameters(a, b, c, alpha, beta, gamma)

    # Extract the basis information
    basis = material_data["basis"]
    elements = [element["value"] for element in basis["elements"]]
    coordinates = [coord["value"] for coord in basis["coordinates"]]

    # Assuming that the basis units are fractional since it's a crystal basis
    coords_are_cartesian = "units" in basis and basis["units"].lower() == "angstrom"

    # Create the Structure
    structure = Structure(lattice, elements, coordinates, coords_are_cartesian=coords_are_cartesian)

    return structure


def from_pymatgen(structure: Structure):
    """
    Converts a pymatgen Structure object to a material object in ESSE format.

    Args:
        structure (Structure): A pymatgen Structure object.

    Returns:
        dict: A dictionary containing the material information in ESSE format.
    """
    basis = {
        "elements": [{"id": i, "value": str(site.specie)} for i, site in enumerate(structure.sites)],
        "coordinates": [{"id": i, "value": list(site.frac_coords)} for i, site in enumerate(structure.sites)],
        "units": "crystal",
        "cell": structure.lattice.matrix.tolist(),
        "constraints": [],
    }

    lattice = {
        "a": structure.lattice.a,
        "b": structure.lattice.b,
        "c": structure.lattice.c,
        "alpha": structure.lattice.alpha,
        "beta": structure.lattice.beta,
        "gamma": structure.lattice.gamma,
        "units": {"length": "angstrom", "angle": "degree"},
        "type": "TRI",
        "vectors": {
            "a": structure.lattice.matrix[0].tolist(),
            "b": structure.lattice.matrix[1].tolist(),
            "c": structure.lattice.matrix[2].tolist(),
            "alat": 1,
            "units": "angstrom",
        },
    }

    material = {
        "name": structure.formula,
        "basis": basis,
        "lattice": lattice,
        "isNonPeriodic": not structure.is_ordered,
        "_id": "",
        "metadata": {"boundaryConditions": {"type": "pbc", "offset": 0}},
        "isUpdated": True,
    }

    return material


def poscar_to_ase(poscar: str):
    """
    Converts POSCAR to ase.Atoms object

    Args:
        poscar (str): POSCAR content

    Returns:
        ase.Atoms: ase.Atoms object
    """
    input_ = io.StringIO(poscar)
    atoms = read(input_, format="vasp")

    return atoms


def ase_to_poscar(atoms: Atoms):
    """
    Converts ase.Atoms object to POSCAR format

    Args:
        atoms (ase.Atoms): ase.Atoms object

    Returns:
        str: POSCAR string
    """
    output = io.StringIO()
    write(output, atoms, format="vasp")
    content = output.getvalue()
    output.close()

    return content


def ase_to_pymatgen(atoms: Atoms):
    """
    Converts ase.Atoms object to pymatgen Structure object

    Args:
        atoms (ase.Atoms): ase.Atoms object

    Returns:
        Structure: pymatgen Structure object
    """

    adaptor = AseAtomsAdaptor()
    structure = adaptor.get_structure(atoms)

    return structure


def pymatgen_to_ase(structure: Structure):
    """
    Converts a Pymatgen Structure object to an ASE Atoms object,
    transferring the 'interface_label' property to the ASE atom tags.

    Args:
        structure (Structure): The Pymatgen Structure object to convert.

    Returns:
        Atoms: The resulting ASE Atoms object with 'interface_label' as tags.
    """
    # Define a mapping from string labels to integer tags
    label_to_tag = {
        "substrate": 0,
        "film": 1,
    }

    adaptor = AseAtomsAdaptor()
    atoms = adaptor.get_atoms(structure)

    # Transfer 'interface_label' property to atom tags using the mapping
    for i, site in enumerate(structure):
        label = site.properties.get("interface_label", None)
        tag = label_to_tag.get(label, 0)  # Default to 0 if 'interface_label' is not set or is unrecognized
        atoms[i].tag = tag

    return atoms


def calculate_average_interlayer_distance(atoms, tag_substrate, tag_film, threshold=0.5):
    """
    Calculate the average distance between the top layer of substrate atoms and the bottom layer of film atoms.

    Args:
        atoms (ase.Atoms): The ASE Atoms object containing both sets of atoms.
        tag_substrate (int): The tag representing the substrate atoms.
        tag_film (int): The tag representing the film atoms.
        threshold (float): The threshold for identifying the top and bottom layers of atoms.

    Returns:
        float: The average distance between the top layer of substrate and the bottom layer of film.
    """
    # Extract z-coordinates of substrate and film atoms
    z_substrate = atoms.positions[atoms.get_tags() == tag_substrate][:, 2]
    z_film = atoms.positions[atoms.get_tags() == tag_film][:, 2]

    # Identify the top layer of substrate atoms and bottom layer of film atoms by z-coordinate
    top_substrate_layer_z = np.max(z_substrate)
    bottom_film_layer_z = np.min(z_film)

    # Get the average z-coordinate of the top substrate atoms (within a threshold from the top)
    top_substrate_atoms = z_substrate[z_substrate >= top_substrate_layer_z - threshold]
    avg_z_top_substrate = np.mean(top_substrate_atoms)

    # Get the average z-coordinate of the bottom film atoms (within a threshold from the bottom)
    bottom_film_atoms = z_film[z_film <= bottom_film_layer_z + threshold]
    avg_z_bottom_film = np.mean(bottom_film_atoms)

    # Calculate the average distance between the top layer of substrate and the bottom layer of film
    average_interlayer_distance = avg_z_bottom_film - avg_z_top_substrate
    return average_interlayer_distance


def filter_atoms_by_tag(atoms: Atoms, material_index):
    """
    Filter atoms by their tag, corresponding to the material index.

    Args:
        atoms (ase.Atoms): The Atoms object to filter.
        material_index (int): The numeric tag to filter by.

    Returns:
        ase.Atoms: The filtered Atoms object.
    """
    return atoms[atoms.get_tags() == material_index]


def get_surface_area(atoms: Atoms):
    """
    Calculate the surface area of the atoms.

    Args:
        atoms (ase.Atoms): The Atoms object to calculate the surface area of.

    Returns:
        float: The surface area of the atoms.
    """
    matrix = atoms.cell
    return np.linalg.norm(np.cross(matrix[0], matrix[1]))


def get_total_energy(atoms: Atoms, calculator: Calculator):
    """
    Set calculator for ASE Atoms and calculate the total energy.

    Args:
        atoms (ase.Atoms): The Atoms object to calculate the energy of.
        calculator (ase.calculators.calculator.Calculator): The calculator to use for the energy calculation.

    Returns:
        float: The total energy of the atoms.
    """
    atoms.set_calculator(calculator)
    return atoms.get_total_energy()


def get_total_energy_per_atom(atoms: Atoms, calculator: Calculator):
    """
        Calculate the energy per atom.

    Args:
            atoms (ase.Atoms): The Atoms object to calculate the energy of.
            calculator (ase.calculators.calculator.Calculator): The calculator to use for the energy calculation.

    Returns:
            float: The energy per atom of the atoms.
    """
    return get_total_energy(atoms, calculator) / atoms.get_global_number_of_atoms()


def get_surface_energy(slab: Atoms, bulk: Atoms, calculator: Calculator):
    """
    Calculate the surface energy by subtracting the weighted bulk energy from the slab energy.

    Args:
        slab (ase.Atoms): The slab Atoms object to calculate the surface energy of.
        bulk (ase.Atoms): The bulk Atoms object to calculate the surface energy of.
        calculator (ase.calculators.calculator.Calculator): The calculator to use for the energy calculation.

    Returns:
        float: The surface energy of the slab.
    """
    number_of_atoms = slab.get_global_number_of_atoms()
    area = get_surface_area(slab)
    return (get_total_energy(slab, calculator) - get_total_energy_per_atom(bulk, calculator) * number_of_atoms) / (
        2 * area
    )


def get_adhesion_energy(interface: Atoms, substrate_slab: Atoms, layer_slab: Atoms, calculator: Calculator):
    """
    Calculate the adhesion energy.
    The adhesion energy is the difference between the energy of the interface and the sum of the energies of the substrate and layer.
    According to: 10.1088/0953-8984/27/30/305004

    Args:
        interface (ase.Atoms): The interface Atoms object to calculate the adhesion energy of.
        substrate_slab (ase.Atoms): The substrate slab Atoms object to calculate the adhesion energy of.
        layer_slab (ase.Atoms): The layer slab Atoms object to calculate the adhesion energy of.
        calculator (ase.calculators.calculator.Calculator): The calculator to use for the energy calculation.

    Returns:
        float: The adhesion energy of the interface.
    """
    energy_substrate_slab = get_total_energy(substrate_slab, calculator)
    energy_layer_slab = get_total_energy(layer_slab, calculator)
    energy_interface = get_total_energy(interface, calculator)
    area = get_surface_area(interface)
    return (energy_substrate_slab + energy_layer_slab - energy_interface) / area


def get_interfacial_energy(
    interface: Atoms,
    substrate_slab: Atoms,
    substrate_bulk: Atoms,
    layer_slab: Atoms,
    layer_bulk: Atoms,
    calculator: Calculator,
):
    """
    Calculate the interfacial energy.
    The interfacial energy is the sum of the surface energies of the substrate and layer minus the adhesion energy.
    According to Dupré's formula

    Args:
        interface (ase.Atoms): The interface Atoms object to calculate the interfacial energy of.
        substrate_slab (ase.Atoms): The substrate slab Atoms object to calculate the interfacial energy of.
        substrate_bulk (ase.Atoms): The substrate bulk Atoms object to calculate the interfacial energy of.
        layer_slab (ase.Atoms): The layer slab Atoms object to calculate the interfacial energy of.
        layer_bulk (ase.Atoms): The layer bulk Atoms object to calculate the interfacial energy of.
        calculator (ase.calculators.calculator.Calculator): The calculator to use for the energy calculation.

    Returns:
        float: The interfacial energy of the interface.
    """

    surface_energy_substrate = get_surface_energy(substrate_slab, substrate_bulk, calculator)
    surface_energy_layer = get_surface_energy(layer_slab, layer_bulk, calculator)
    adhesion_energy = get_adhesion_energy(interface, substrate_slab, layer_slab, calculator)
    return surface_energy_layer + surface_energy_substrate - adhesion_energy


def from_ase(atoms: Atoms):
    """
    Convert an ASE Atoms object to a material object in ESSE format.

    Args:
        atoms (ase.Atoms): The ASE Atoms object to convert.

    Returns:
        dict: A dictionary containing the material information in ESSE format.
    """
    structure = ase_to_pymatgen(atoms)
    material = from_pymatgen(structure)
    return material


def translate_to_bottom(structure):
    """
    Translate the structure to the bottom of the cell.
    Args:
        structure (Structure): The pymatgen Structure object to translate.

    Returns:
        Structure: The translated pymatgen Structure object.
    """
    min_c = min(site.c for site in structure)
    translation_vector = [0, 0, -min_c]
    translated_structure = structure.copy()
    for site in translated_structure:
        site.coords += translation_vector
    return translated_structure



#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/job/get-file-from-job.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Get-File-From-Job
#
# This example demonstrates how to use Mat3ra RESTful API to check for and acquire files from jobs which have been run. This example assumes that the user is already familiar with the [creation and submission of jobs](create_and_submit_jobs.ipynb) using our API.
#
# > <span style="color: orange">**IMPORTANT NOTE**</span>: In order to run this example in full, an active Mat3ra.com account is required. Alternatively, Readers may substitute the workflow ID below with another one (an equivalent one for VASP, for example) and adjust extraction of the results ("Viewing job files" section). RESTful API credentials shall be updated in [settings](../../utils/settings.json).
#
#
# ## Steps
#
# After working through this notebook, you will be able to:
#
# 1. Import [the structure of Si](https://materialsproject.org/materials/mp-149/) from Materials Project
# 2. Set up and run a single-point calculation using Quantum Espresso.
# 3. List files currently in the job's directory
# 4. Check metadata for every file (modification date, size, etc)
# 5. Access file contents directly and print them to console
# 6. Download files to your local machine
#
# ## Pre-requisites
#
# The explanation below assumes that the reader is familiar with the concepts used in Mat3ra platform and RESTful API. We outline these below and direct the reader to the original sources of information:
#
# - [Generating RESTful API authentication parameters](../system/get_authentication_params.ipynb)
# - [Importing materials from materials project](../material/import_materials_from_materialsproject.ipynb)
# - [Creating and submitting jobs](../job/create_and_submit_job.ipynb)

# # Complete Authorization Form and Initialize Settings
#
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
#
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
#
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
#
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
#
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
#
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# ## Imports

# In[ ]:


# Import settings file and utils file
from utils.settings import ENDPOINT_ARGS, ACCOUNT_ID, MATERIALS_PROJECT_API_KEY
from utils.generic import (
    wait_for_jobs_to_finish,
    get_property_by_subworkflow_and_unit_indicies,
    dataframe_to_html,
    display_JSON,
)

# Relevant functions from the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints
from exabyte_api_client.endpoints.properties import PropertiesEndpoints


# ### Create and submit the job
#
# For this job, we'll use the workflow located [here](https://platform.mat3ra.com/analytics/workflows/84DAjE9YyTFndx6z3).
#
# This workflow is a single-point total energy calculation using Density-Functional Energy as-implemented in Quantum Espresso version 5.4.0.
#
# The PBE functional is used in conjunction with an ultrasoft pseudopotential and a planewave basis set.
#
# The material we will investigate is elemental [Silicon](https://materialsproject.org/materials/mp-149/), as-is from Materials Project.
#
# > <span style="color: orange">Note</span>: This cell uses our API to copy the unit cell of silicon from Materials Project into your account. It then copies a workflow to get the total energy of a system using Quantum Espresso to your account. Finally, a job is created using the Quantum Espresso workflow for the silicon unit cell, and the job is submitted to the cluster. For more information, please refer to our [run-simulation-and-extract-properties](./run-simulations-and-extract-properties.ipynb) notebook, located in this directory.

# In[ ]:


# Get some account information
project_endpoints = ProjectEndpoints(*ENDPOINT_ARGS)
project_metadata = project_endpoints.list({"isDefault": True, "owner._id": ACCOUNT_ID})[0]
project_id = project_metadata["_id"]
owner_id = project_metadata["owner"]["_id"]

# Get a workflow for the job from the bank, and copy it to our account
bank_workflow_endpoints = BankWorkflowEndpoints(*ENDPOINT_ARGS)
BANK_WORKFLOW_ID = "84DAjE9YyTFndx6z3"
workflow_id = bank_workflow_endpoints.copy(BANK_WORKFLOW_ID, owner_id)["_id"]

# Get materials for the job
material_endpoints = MaterialEndpoints(*ENDPOINT_ARGS)
material_project_id = ["mp-149"]  # The importer expects a list
materials = material_endpoints.import_from_materialsproject(MATERIALS_PROJECT_API_KEY, material_project_id, owner_id)

# Create the job
job_endpoints = JobEndpoints(*ENDPOINT_ARGS)
job = job_endpoints.create_by_ids(
    materials=materials, workflow_id=workflow_id, project_id=project_id, owner_id=owner_id, prefix="Test_Job_Output"
)[0]

# Submit the job
job_endpoints.submit(job["_id"])
wait_for_jobs_to_finish(job_endpoints, [job["_id"]])


# Monitor the jobs and print the status until they are all finished.

# ## Viewing job files
# ### Retreive a list of job files
#
# Here, we'll get a list of all files that belong to the job.

# In[ ]:


files = job_endpoints.list_files(job["_id"])
paths = [file["key"] for file in files]
for path in paths:
    if "outdir" not in path:
        print(path)


# ### Get metadata for the Output File
# The .out file is where Quantum Espresso shows its work and prints its results, so you most likely will want to view this files. Let's print out some of its metadata.
#
# You'll find that we get a lot of data describing the file and its providence. Brief explanations of each entry are:
# - Key - Path to the file on the cluster
# - size - Size of the file, in bytes.
# - Bucket - The name of the cluster which ran the job.
# - Region - Which server region was used to run the job.
# - Provider - The cluster provider for the compute resources (in our case, we used AWS).
# - lastModified - Unix timestamp representing when the file was last modified.
# - name - The filename.
# - signedUrl - This is a link which can be used to download the file for a short amount of time.

# In[ ]:


for file in files:
    if file["name"] == "pw_scf.out":
        output_file_metadata = file
display_JSON(output_file_metadata)


# ### Display file contents to console
#
# The signedUrl gives us a place to access the file and download it. Let's read it into memory, and print out the last few lines of our job.

# In[ ]:


import urllib

server_response = urllib.request.urlopen(output_file_metadata["signedUrl"])
output_file_bytes = server_response.read()

# The server returns us a bytes-string. That's useful for things like binaries or other non-human-readable data, but this should be decoded if we're planning to write to console.
# Because this is a human-readable text file, we'll decode it to UTF-8.
output_file = output_file_bytes.decode(encoding="UTF-8")

# Tail the last 90 lines
lines = output_file.split("\n")
for line in lines[-90:]:
    print(line)


# ### Save the input file and output file to disk.
#
# Now that we've verified the job is done, let's go ahead and save it and its input to disk.

# In[ ]:


# We've already got an output file, so let's grab the input file we sent to Quantum Espresso
for file in files:
    if "pw_scf.in" == file["name"]:
        input_file_metadata = file
server_response = urllib.request.urlopen(input_file_metadata["signedUrl"])
input_file_bytes = server_response.read()


# In[ ]:


# Let's write the input file to disk. Note that we get files as a bytes string from the server, which is convenient for binaries, images, and other non-human-readable data.
# Although we could decode before writing to disk, we can just write it directly with the "wb" (write bytes) file mode.
with open(input_file_metadata["name"], "wb") as file_descriptor:
    file_descriptor.write(input_file_bytes)


# In[ ]:


# Now, let's write our output file to the disk. Note that because we already decoded it, we can just use the 'w' file mode.
with open(output_file_metadata["name"], "w") as file_descriptor:
    file_descriptor.write(output_file)




#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/material/create_material.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Overview
# 
# In this example we create a material from a JSON config with [tags](https://docs.mat3ra.com/entities-general/data/#tags) to identify the material.

# # Complete Authorization Form and Initialize Settings
# 
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
# 
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
# 
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
# 
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
# 
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
# 
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# # Imports

# In[ ]:


from utils.settings import ENDPOINT_ARGS
from utils.generic import display_JSON

from exabyte_api_client.endpoints.materials import MaterialEndpoints


# ## Create material config
# 
# Create material config in JSON format. See [Material](https://docs.mat3ra.com/api/Material/put_materials_create) endpoint for more information about material config format.

# In[ ]:


CONFIG = {
    "name": "TEST MATERIAL",
    "basis": {
        "elements": [{"id": 1, "value": "Si"}, {"id": 2, "value": "Si"}],
        "coordinates": [{"id": 1, "value": [0, 0, 0]}, {"id": 2, "value": [0.25, 0.25, 0.25]}],
        "units": "crystal",
        "name": "basis",
    },
    "lattice": {
        "type": "FCC",
        "a": 3.867,
        "b": 3.867,
        "c": 3.867,
        "alpha": 60,
        "beta": 60,
        "gamma": 60,
        "units": {"length": "angstrom", "angle": "degree"},
        "vectors": {
            "a": [3.867, 0, 0],
            "b": [1.9335000000000004, 3.348920236434424, 0],
            "c": [1.9335000000000004, 1.1163067454781415, 3.1573922784475164],
            "name": "lattice vectors",
            "alat": 1,
            "units": "angstrom",
        },
    },
    "tags": ["REST API"],
}


# ## Create material
# 
# Initialize `MaterialEndpoints` class and call `create` function to create material.

# In[ ]:


endpoint = MaterialEndpoints(*ENDPOINT_ARGS)
material = endpoint.create(CONFIG)


# ## Print new material

# In[ ]:


display_JSON(material)




# This script is used to serve the static files from the 'dist' directory for installation of python wheels.
# 1. Install deps with `pip install ".[dev]"`,
# 2. Install build tool with `pip install build`
# 3. Run `python -m build`
# You should see something alongside:
# ```Successfully built mat3ra_api_examples-dev9+g7c6e8d9.tar.gz and
# mat3ra_api_examples-dev9+g7c6e8d9-py3-none-any.whl
# ```
# 4. Copy the wheel file name
# 5. Run the server with `python wheel_server.py`
# The server will be available at http://localhost:8080,
# 6. use micropip in pyodide to install wheel from that URL `<server_url>/<wheel_file_name>`, i.e:
# await micropip.install("http://localhost:8080/mat3ra_api_examples-dev9+g7c6e8d9-py3-none-any.whl", deps=False)

import argparse
import glob
import os
import socket
from http.server import HTTPServer, SimpleHTTPRequestHandler


class CORSHTTPRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self):
        self.send_header("Access-Control-Allow-Origin", "*")
        self.send_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
        self.send_header("Access-Control-Allow-Headers", "X-Requested-With, Content-Type")
        return super(CORSHTTPRequestHandler, self).end_headers()


def check_port(host, port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def inform_user(port):
    whl_files = glob.glob("*.whl")
    file = whl_files[0] if whl_files else None
    url_str = f"http://localhost:{port}/{file}"
    print("Copy URL to use in notebook or `config.yml`: ", url_str)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Python wheel server.")
    parser.add_argument("--port", type=int, default=8080, help="Port to run the server on.")
    parser.add_argument("--dir", type=str, default="./dist", help="Directory to serve.")
    args = parser.parse_args()

    port = args.port
    bind_addr = "localhost"
    directory = args.dir  # Change this line

    os.chdir(directory)  # Change the current working directory to the specified 'directory'

    while check_port(bind_addr, port):
        print(f"Port {port} is already in use. Trying with port {port + 1}.")
        port += 1

    httpd = HTTPServer((bind_addr, port), CORSHTTPRequestHandler)
    print(f"Serving at http://{bind_addr}:{port}")
    inform_user(port)
    httpd.serve_forever()



from typing import Dict, List, Union

import plotly.graph_objs as go
from ase.atoms import Atoms as ASEAtoms
from ase.optimize import BFGS, FIRE
from IPython.display import display
from mat3ra.made.material import Material
from mat3ra.made.tools.build.interface.enums import StrainModes
from plotly.subplots import make_subplots


def plot_strain_vs_atoms(interfaces: List[Material], settings: Dict[str, Union[str, int]]):
    """
    Plot the strain vs the number of atoms for each interface.
    Allows for visual inspection to find the most optimal interface for current task.
    Args:
        interfaces (List[Material]): The interfaces to plot.
        settings (Dict[str, Union[str, int]]): The settings for the plot.
    """

    data = []
    for index, interface in enumerate(interfaces):
        strain_percentage = interface.metadata[StrainModes.mean_abs_strain] * 100
        num_sites = len(interface.basis.coordinates.values)

        hover_text = f"Index: {index}<br>Strain: {strain_percentage:.2f}%<br>Atoms: {num_sites}"

        trace = go.Scatter(
            x=[strain_percentage],
            y=[num_sites],
            text=[hover_text],
            mode="markers",
            hoverinfo="text",
            name=f"Index: {index}",
        )
        data.append(trace)

    layout = go.Layout(
        xaxis=dict(title="Strain (%)", type=settings["X_SCALE"]),
        yaxis=dict(title="Number of atoms", type=settings["Y_SCALE"]),
        hovermode="closest",
        height=settings["HEIGHT"],
        legend_title_text="Interfaces Indices",
    )

    fig = go.Figure(data=data, layout=layout)
    fig.show()


def create_realtime_plot():
    """
    Create a real-time plot for optimization progress.
    Returns:
        go.FigureWidget: The real-time plot.
    """
    fig = make_subplots(rows=1, cols=1, specs=[[{"type": "scatter"}]])
    scatter = go.Scatter(x=[], y=[], mode="lines+markers", name="Energy")
    fig.add_trace(scatter)
    fig.update_layout(title_text="Real-time Optimization Progress", xaxis_title="Step", yaxis_title="Energy (eV)")
    f = go.FigureWidget(fig)
    display(f)
    return f


def plot_update_callback(
    dynamic_object: Union[BFGS, FIRE],
    ase_interface: ASEAtoms,
    plotly_figure: go.FigureWidget,
    steps: List[int],
    energies: List[int],
):
    """
    Callback function for updating energies for steps in real-time.
    Args:
        dynamic_object: The ASE dynamics object.
        ase_interface: The ASE interface object.
        plotly_figure: The plotly figure widget.
        steps: The list of steps.
        energies: The list of energies.

    Returns:
        function: The update function that is attached to the dynamics object and called each step.
    """

    def update():
        step = dynamic_object.nsteps
        energy = ase_interface.get_total_energy()

        steps.append(step)
        energies.append(energy)

        print(f"Step: {step}, Energy: {energy:.4f} eV")
        with plotly_figure.batch_update():
            plotly_figure.data[0].x = steps
            plotly_figure.data[0].y = energies

    return update



#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/system/get_authentication_params.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Overview
# 
# This example shows how to log in to Mat3ra RESTFul API via [Login](https://docs.mat3ra.com/api/API/post_login) endpoint and generate API authentication parameters. Note that it is also possible to generate an API token manually via platform web UI as described in the project [README](https://github.com/Exabyte-io/api-examples/tree/dev#usage).

# # Execution
# 
# > <span style="color: orange">**NOTE**</span>: In order to run this example, an active Mat3ra.com account is required.

# ## Set Parameters
# 
# - **USERNAME**: Your Mat3ra account username.
# 
# - **PASSWORD**: Your Mat3ra account password.

# In[ ]:


# @title Authorization Form
USERNAME = "YOUR_USERNAME"  # @param {type:"string"}

# avoid storing password in plaintext
from getpass import getpass

PASSWORD = getpass("Please enter password: ")

import os

if "COLAB_JUPYTER_IP" in os.environ:
    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# ## Import packages

# In[ ]:


from utils.settings import HOST, PORT, VERSION, SECURE
from utils.generic import display_JSON

from exabyte_api_client.endpoints.login import LoginEndpoint


# ## Initialize the endpoint

# In[ ]:


endpoint = LoginEndpoint(HOST, PORT, USERNAME, PASSWORD, VERSION, SECURE)
auth_params = endpoint.login()


# ## Print authentication parameters
# 
# Print the authentication parameters in pretty JSON below. Update [settings](../../utils/settings.json) with this parameters to be able to run other examples.

# In[ ]:


display_JSON(auth_params)




#!/usr/bin/env python
# coding: utf-8

# # Background
# 
# This notebook walks the user though the design and tuning of a neural network to predict the adsorption energetics of $\cdot{CH_3}$, $\cdot{OH}$, and $CO$ (several key intermediates in a variety of catalytic processes including $CO_2$ hydrogenation) to nanoparticles made of Cu, Ag, and Au. 
# 
# ## Adsorption Energetics
# 
# Broadly, this notebook re-implements a recent adsorption model found by Dean et al and published in [Science Advances](https://advances.sciencemag.org/content/5/9/eaax5101.abstract). This work was motivated by the need to predict the adsorption energetics of molecular species to heterogeneous catalysts. Understanding this physical property is important, as catalytic activity is [often a function](https://en.wikipedia.org/wiki/Sabatier_principle) of the adsorption strength of key intermediates, and tuning a catalysts's adsorption strength can be a key part of optimizing its catalytic activity. Typically, one calculates adsorption energy by performing three separate Density-Functional Theory (DFT) calculations:
# 
# 1. The adsorbate (for example, CO)
# 2. The catalyst (for example, a nanoparticle)
# 3. The adsorbed state (for example, CO adsorbed to a particular adsorption site on a nanoparticle)
# 
# The adsorption energy is then calculated as follows, where $E_{ads}$ is the adsorption energy, $E_{adsorbate}$ is the energy of the adsorbate (calculation #1 above), $E_{catalyst}$ is the energy of the catalyst (calculation #2 above), and $E_{adsorbate-catalyst}$ is the energy of the complex between the two (calculation #3 above):
# 
# $E_{ads} = E_{adsorbate-catalyst} - E_{adsorbate} - E_{catalyst}$
# 
# In simpler terms, this equations calculates the energy required to begin at the adsorbed state, and separate the adsorbate and catalyst at an infinite distance from one-another.
# 
# This can become computationally expensive, as it requires three different DFT calculations for one material. Hence, this motivates the model created by [Dean et al](https://advances.sciencemag.org/content/5/9/eaax5101.abstract), as the creation of a computationally-inexpensive method of predicting adsorption energetics can increase the throughput of catalyst screening, thus accelerating catalyst discovery.
# 
# ## The Training Set
# 
# The training set consists of the adsorption energetics to sites on various nanoparticles made of copper, silver, and gold (metals which are commonly investigated for catalytic applications). Adsorbates investigated were $\cdot{CH_3}$, $\cdot{OH}$, and $CO$, which were chosen because of their relevance as intermediates in various catalytic cycles. Data was generated using [CP2K](https://www.cp2k.org/)'s implementation of DFT. A Double-Zeta Valence Polarized (DZVP) basis set was used in conjunction with the pseudopotentials of Goedecker, Teter, and Hutter (with a 500Ry cutoff) and the Perdew-Burke-Ernzerhof (PBE) functional. Overall, this results in a robust methodology which is generally good at capturing the adsorption interaction of catalytically-relevant species to transition metals.
# 
# ## Descriptors Used
# 
# In the work, three desriptors were found: $CE_{local}$, $\mu_{adsorbate}$, and MADs. They were chosen for their rapid determination, either being tabulated or computationally-inexpensive to determine.
# 
# - $CE_{local}$ is a descriptor which represents the stability of the binding site. Specifically, if we consider a metal atom bound to a nanoparticle, $CE_{local}$ is equal to the sum of bond energies for every bond it forms with the cluster. For example, if atom A was bound to atoms B, C, and D in a nanoparticle, $CE_{local}$ would be equal to the sum of the A-B bond energy, the A-C bond energy, and the A-D bond energy. Implementation-wise, to avoid requiring DFT calculations, Dean et al utilized a variation of the [Bond-Centric Model](https://pubs.acs.org/doi/abs/10.1021/acs.nanolett.8b00670) (BCM) of nanoparticle stability, which enables the rapid prediction of bond energies in a metallic system.
# 
# - $\mu_{adsorbate}$ is the chemical potential of the adsorbate, which can be approximated under Hard/Soft Acid/Base (HSAB) theory via a first-order central differencing approximation using the Electron Affinity (EA) and Ionization Potential (IP) of the molecule. This results to the negative average of the IP and EA. For many relevant chemical species, the IP and EA area readily tabulated on resources such as the NIST WebBook or CRC Handbook of Chemistry.
# 
# - MADs functions as a descriptor of the intrinsic tendency of the <u>M</u>etal and <u>ADS</u>orbate to bind. It is the binding energy between the adsorbate and a single metal atom of the relevant metal type (for example, for CO adsorbing to Cu, it would be the binding strength of a single CO molecule adsorbing to a single atom of Cu). Although this is performed via DFT, there are few-enough atoms that this calculation is highly tractable. Moreover, because it is only performed once for a given adsorbate-metal pair, it does not present a bottleneck to a high-throughput screening approach.
# 
# ## Neural Networks
# 
# In the original paper, a linear regression is found using the above descriptors with a 10-fold cross validated Root Mean Squared Error (RMSE) of 0.179 eV:
# 
# $E_{ads} \approx a*CE_{local} + b*\mu_{adsorbate} + c*MADs + d$
# 
# In this notebook, we replicate the results of this linear regression using a neural network. We proceed through a -step process:
# 
# 1. We import the adsorption data along with the above descriptors descriptor (available in the supporting information of [the paper](https://advances.sciencemag.org/content/5/9/eaax5101.abstract))
# 2. We create a function to programatically create a neural network using TensorFlow
# 3. We optimze several neural network hyperparameters using Scikit-Learn's cross-validated grid search implementation
# 4. We train a final neural network using the tuned hyperparameters, and evaluate it by:
#     a. Learning curves for the train / test set
#     b. A parity plot of predicted versus ground-truth adsorption energy

# # Setting Up

# In[]:


import random
import logging
import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn.model_selection
import subprocess
import tensorflow as tf
import IPython
from tensorflow.keras.layers import Dense, BatchNormalization, InputLayer
from tensorflow.keras.wrappers.scikit_learn import KerasRegressor


# In[]:


# Random seeds for reproducibility
np.random.seed(1234)
random.seed(1234)


# Next, we'll get everything into a Pandas dataframe.

# In[]:


# Load data into a Pandas dataframe for convenience
df = pd.read_csv("../assets/sci_adv_dean_data.csv", sep=",")
y_cols = ["PBE_BE_eV"]
x_cols = ["CE_Local_eV", "ChemPot_eV", "MADS_eV"]

# Split up into X/Y values, for convenience
train_x = df[x_cols]
train_y = df[y_cols]

# Display a friendly CSV of the data within Jupyter.
IPython.display.display(df)


# ## Hyperparameter Optimization
# 
# The "Hyperparameter Optimization" step of a typical ML process involves tuning various parts of the machine learning implementation. For example, in a neural network, one hyperparameter is the so-called "architecture" of the network. In other words, where the different neurons are, how many neurons there are, the activation functions used in each neuron, etc. In our case, to keep things simple, we only tune the depth and width of the network.
# 
# Next, we create a function to automatically create a TF feedforward neural network with a given width an depth. We then use this function alongside TF's wrapper for the Sklearn regressor API to produce a build function for use in the grid search.

# In[]:


def create_model(hidden_layers_width: int, hidden_layers_depth: int) -> tf.keras.Sequential:
    """
    Function to programatically create a standard feedforward neural network with a given architecture.
    
    Args:
        hidden_layers_width (int): How many neurons are inside each layer.
        hidden_layers_depth (int): How many hidden layers the neural network has.
        
    Returns:
        A keras Sequential model object.
    """
    
    # Initialize the model object
    model = tf.keras.Sequential()
    
    # Input layer of the model
    model.add(InputLayer(input_shape=len(x_cols),
                         name="Input_Layer")
             )
    
    # Batch normalization is useful, normalizing the data before it continues deeper into the network.
    model.add(BatchNormalization(name="Batch_Normalization"))
    
    # Hidden layers
    # These are typically important for ensuring the network has enough freedom to tackle the regression problem.
    # A neural network without a hidden layer is typically called a "Perceptron," and is very limited in the types of data it can model.
    for layer in range(hidden_layers_depth):
        model.add(Dense(hidden_layers_width,
                        activation="relu",
                        name=f"Hidden_Layer_{layer+1}")
                 )
        
    # Output layer
    # Because this is a regression problem, here we use linear activation (e.g. summing every input plus the activation)
    model.add(Dense(1, # Only one output neuron
                    activation="linear",
                    name="Output_Layer")
             )
    
    # Compile the model object and return it
    model.compile(loss="mse")
    return model

# TensorFlow / Keras comes with a convenient wrapper to facilitate its integration it the Scikit-Learn framework.
nn_model = KerasRegressor(build_fn=create_model,
                          verbose=0,    # Ranges from 0-2; how much output the model should print while it is training.
                          epochs=256,   # How many epocs (number of complete passes over the training set) to train the network.
                          batch_size=16 # How many examples to consider at a time (See https://towardsdatascience.com/batch-mini-batch-stochastic-gradient-descent-7a62ecba642a for a detailed explanation)
                          )


# We now conduct a grid search over several possible architectures.

# In[]:


# Number of folds for cross-validation
folds = 5

# How many threads to use for the grid search
n_threads = 4

# Hidden layer widths to check
hidden_layer_widths = [3]

# Depths to check
hidden_layer_depths = [0, 1, 2, 4, 8]

# Put together the dictionary of tuning parameters, for our GridSearchCV later on
tuning_parameters = {"hidden_layers_width": hidden_layer_widths,
                     "hidden_layers_depth": hidden_layer_depths}


# In[]:


print(f"Training on {len(train_x)} samples across {len(train_x.columns)} features")
print(f"{folds}-fold CV")

# Set up the grid
grid = sklearn.model_selection.GridSearchCV(estimator=nn_model,               # SKlearn Regressor/Classifier object to tune. This object gets repeatedly trained.
                                            param_grid=tuning_parameters,     # Which parameters to check (in our case, width and depth). All possibilit combos will be checked.
                                            scoring="neg_mean_squared_error", # How to rank each combination of hyperparameters (here, we use RMSE). Note that GridSearchCV tries to minimize the scoring metric, hence we use the negative version of RMSE.
                                            verbose=1,                        # Ranges from 1-3. How much information to print to the screen during the grid search.
                                            cv=folds,                         # Controls the number of folds to be used in cross-validation.
                                            n_jobs=n_threads                  # How many models are to be trained simultaneously.
                                            )

# Ensure TensorFlow stays silent during training ( to avoid flooding the console with warning messages, since TF occasionally uses depreciated features of NumPy )
logger = tf.get_logger()
logger.setLevel(logging.ERROR)

# Fit the grid
grid.fit(X=train_x, y=train_y)


# Now, let's print out some information about the best architecture found:

# In[]:


# Print the RMSE
rmse = np.round(np.sqrt(abs(grid.best_score_)),3)
print(f"RMSE = {rmse}")

# Print the best set of hyperparameters which were found
best_params = grid.best_params_
print("\nBest parameters are:")
for key, value in best_params.items():
    print(f"\t{key} : {value}")

# Using this set of hyperparameters, compile the final form of the model.
best_model = create_model(hidden_layers_width=best_params["hidden_layers_width"],
                          hidden_layers_depth=best_params["hidden_layers_depth"])

# Print out detailed information regarding the neural network architecture.
#     "Output Shape" refers to the shape of the vector/tensor the given layer sends to the next.
#     "Param #" refers to the number of parameters involved in the creation of that layer (i.e. weights, biases, etc). This is analogous to the concept of "Degrees of Freedom" in least-squares regression.
print("\nDetailed Information:")
print(best_model.summary())

# Draw a graph of the architecture below (this is a useful feature for drafting paper/presentation figures!)
print("\nArchitecture:")
# GraphViz (specifically, its executable named "dot") does need to be in the system path for this to work. Here, we check that it's installed.
try:
    subprocess.call(['dot', '-v'])
    dot_exists = True
except FileNotFoundError:
    dot_exists = False
# Now that we know whether it exists, we might draw the network.
# Note that this will create a new file named "architecture.png" (or whatever arch_filename is assigned to) in the current working directory
if dot_exists:
    arch_filename = "architecture.png"
    tf.keras.utils.plot_model(best_model, show_shapes=True, to_file=arch_filename)
    image = IPython.display.Image(filename=arch_filename)
    IPython.display.display(image)
else:
    # If GraphViz is not in the system path, we'll error out if we try to use it.
    print("dot is not found in the System path, but was needed to draw the architecture.")


# ## Final Training
# 
# Now that we have identified a good model architecture, we train it.

# In[]:


# Train the final network
history = best_model.fit(train_x, train_y,                      # Training Data
                         epochs=100000,                         # Maximum number of epochs to use. Note that here, we use the EarlyStopping callback to halt training once the error has stopped improving.
                         batch_size=16,                         # Number of examples to consider simultaneously in the approximation of the error gradient. See: https://towardsdatascience.com/batch-mini-batch-stochastic-gradient-descent-7a62ecba642a
                         verbose=0,                             # How much to print to console during training. Ranges from 0-2.
                         validation_split=0.1,                  # The percentage of data to leave out as a validation set (in this case, 0.1 leaves out 10%, and trains on the remaining 90%).
                         callbacks=[                            # "Callbacks" are special helper-functions that Keras can use during training. See: https://keras.io/api/callbacks/
                             tf.keras.callbacks.EarlyStopping(  # EarlyStopping halts the training process once the quality of the neural network's predictions has stopped improving.
                                 monitor="val_loss",            # The parameter we're monitoring for improvement. In this case, it's loss of the validation data we've left out of the training set (controlled by validation_split above).
                                 min_delta=0,                   # The minimum amount of change in the error gradient for us to consider it an improvement. In this case, to simplify things we just check whether any improvement has been made at all.
                                 patience=50,                   # How many epochs are allowed to pass without improvement before the network's training is stopped.
                                 mode="min",                    # Whether we are looking for loss to increase or decrease. In our case, we want it to decrease.
                                 restore_best_weights=True)     # This option allows us to restore the parameters which yielded the best model found during the training process.
                         ]
                         ).history # History objects are created when a Keras neural network gets fit.


# ## Final Evaluation
# 
# Now that we have a trained model, we'll plot its learning curve and create a parity plot.

# In[]:


# Parameters for PyPlot. Controls how large the figures are when printed to the screen or saved.
plt.rcParams['figure.figsize'] = 5,5
plt.rcParams['figure.dpi'] = 200


# In[]:


loss = np.sqrt(history["loss"]) # Loss over time of the training set
val_loss = np.sqrt(history["val_loss"]) # Loss over time of the validation set

# Plot the training history
plt.plot(loss, label="Training Set Loss")
plt.plot(val_loss, label="Validation Set Loss")

# Make the plot look a bit better, add labels
plt.legend()
plt.ylabel("RMSE Loss (eV)")
plt.xlabel("Epoch Number")

# Save the plot to disk
fig = plt.gcf()
fig.savefig("learning_curve.png")

# Show the plot to the user
plt.show()
plt.close()


# In[]:


# Get predictions for every row in the dataset
df["predicted_adsorption_eV"] = best_model.predict(train_x)

# Make a parity plot. That is, plot the predictions of the model versus the training set values (which are from DFT)
plt.scatter(x=df["predicted_adsorption_eV"], y=train_y,
            label=f"Predictions (RMSE = {np.round(rmse, 2)} eV)",
            c="#1c2957")

# To make the plot have the same dimensions in X and Y, find the min and max in each dimension.
lims_min = min(min(df["predicted_adsorption_eV"]),
               float(train_y.min())
               )
lims_max = max(max(df["predicted_adsorption_eV"]),
                float(train_y.max())
               )
lims = (lims_min, lims_max)

# Set the limits to ensure the plot is square
plt.xlim(lims)
plt.ylim(lims)

# Plot a line along y=x as a guide to the eye of what perfect parity looks like
plt.plot([lims[0], lims[1]],
         [lims[0], lims[1]],
         label="Parity", c="black")

# Make the plot look a bit better, add labels
plt.legend()
plt.ylabel("DFT Adsorption Energy (eV)")
plt.xlabel("Predicted Adsorption Energy (eV)")

# Save the plot to disk
fig = plt.gcf()
fig.savefig("parity.png")

# Show the plot to the user
plt.show()
plt.close()


# ## Save Model
# 
# And last, we will save the final trained model.

# In[]:


# Save the final model to disk
# This is a very convenient feature of Keras, as the model could then be loaded and used to predict new values later-on, without needing to re-train.
# For a model that's simple like ours, avoiding retraining isn't a huge concern, but for larger networks that might be resource-intensive to train, this could be an important final step.
print(best_params)
best_model.save("best_model.mdl")




#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/job/ml-train-model-predict-properties.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Overview
#
# This example demonstrates how to use Mat3ra RESTful API to build a machine learning (ML) model for a set of materials called "train materials" and use the model to predict properties of another set called "target materials". The general approach can work for multiple properties, we use the Electronic Band Gap in this example.
#
#
#
# ## Steps
#
# We follow the below steps:
#
# - Import materials from [materials project](https://materialsproject.org/)
# - Calculate band gap for the "train materials"
# - Build ML Train model based on the "train materials"
# - Create and submit a job to predict band gap for the "target materials"
# - Extract band gap for "target materials"
# - Output the results as Pandas dataFrame
#
# ## Pre-requisites
#
# The explanation below assumes that the reader is familiar with the concepts used in Mat3ra platform and RESTful API. We outline these below and direct the reader to the original sources of information:
#
# - [Generating RESTful API authentication parameters](../system/get_authentication_params.ipynb)
# - [Importing materials from materials project](../material/import_materials_from_materialsproject.ipynb)
# - [Creating and submitting jobs](./create_and_submit_job.ipynb)
# - [Running DFT calculations](./run-simulations-and-extract-properties.ipynb)

# # Complete Authorization Form and Initialize Settings
#
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
#
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
#
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
#
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
#
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
#
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# # Imports

# In[ ]:


import time
from IPython.display import IFrame

# Import settings file and utils file
from utils.settings import ENDPOINT_ARGS, ACCOUNT_ID, MATERIALS_PROJECT_API_KEY
from utils.generic import (
    dataframe_to_html,
    copy_bank_workflow_by_system_name,
    wait_for_jobs_to_finish,
    get_property_by_subworkflow_and_unit_indicies,
    display_JSON,
)

import pandas as pd

# Import relevant portions of the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.utils.materials import flatten_material
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.workflows import WorkflowEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints
from exabyte_api_client.endpoints.properties import PropertiesEndpoints


# #### Materials
#
# Set parameters for the materials to be imported:
#
# - **TRAIN_MATERIALS_PROJECT_IDS**: a list of material IDs to train ML model based on
# - **TARGET_MATERIALS_PROJECT_IDS**: a list of material IDs to predict the property for

# In[ ]:


TRAIN_MATERIALS_PROJECT_IDS = ["mp-149", "mp-978534"]  # Si, SiGe
TARGET_MATERIALS_PROJECT_IDS = ["mp-32"]  # Ge


# #### Jobs
#
# Set parameters for the jobs to be ran for the imported materials:
#
# - **JOB_NAME_PREFIX**: prefix to be used for the job name with "{JOB_NAME_PREFIX} {FORMULA}" convention (e.g.  "Job Name Prefix - SiGe")

# In[ ]:


JOB_NAME_PREFIX = "Job Name Prefix"


# #### Compute
#
# Setup compute parameters. See [this](https://docs.mat3ra.com/infrastructure/compute-settings/ui) for more information about compute parameters.
#
# - **NODES**: Number of nodes. Defaults to 1.
# - **PPN**: Number of MPI processes per each node, Defaults to 1.
# - **QUEUE**: The name of queue to submit the jobs into. Defaults to D.
# - **TIME_LIMIT**: Job walltime. Defaults to "01:00:00" (one hour).
# - **CLUSTER**: The full qualified domain name (FQDN) or alias of the cluster to submit the jobs into.

# In[ ]:


PPN = "1"
QUEUE = "D"
NODES = "1"
TIME_LIMIT = "01:00:00"
CLUSTER = "cluster-001"


# ### Initialize the endpoints

# In[ ]:


job_endpoints = JobEndpoints(*ENDPOINT_ARGS)
project_endpoints = ProjectEndpoints(*ENDPOINT_ARGS)
material_endpoints = MaterialEndpoints(*ENDPOINT_ARGS)
workflow_endpoints = WorkflowEndpoints(*ENDPOINT_ARGS)
bank_workflow_endpoints = BankWorkflowEndpoints(*ENDPOINT_ARGS)
property_endpoints = PropertiesEndpoints(*ENDPOINT_ARGS)


# Retrieve the owner and project IDs as they are needed by the endpoints. The default material is used to extract the owner ID. One can extract the owner ID from any other account's [entities](https://docs.mat3ra.com/entities-general/overview/).

# In[ ]:


owner_id = material_endpoints.list({"isDefault": True, "owner._id": ACCOUNT_ID})[0]["owner"]["_id"]
project_id = project_endpoints.list({"isDefault": True, "owner._id": ACCOUNT_ID})[0]["_id"]


# ### Create workflows
#
# Copy "ML: Train Model" and "Band Gap" bank workflows to the account's workflows. We use exabyte bank workflows which are identified by "systemName" field. The below can be adjusted to get the bank workflows by ID.

# In[ ]:


band_gap_workflow_id = copy_bank_workflow_by_system_name(bank_workflow_endpoints, "espresso-band-gap", owner_id)
ml_train_workflow_id = copy_bank_workflow_by_system_name(bank_workflow_endpoints, "exabyteml-ml-train-model", owner_id)


# ### Import materials
#
# Import materials from materials project.

# In[ ]:


train_materials = material_endpoints.import_from_materialsproject(
    MATERIALS_PROJECT_API_KEY, TRAIN_MATERIALS_PROJECT_IDS, owner_id
)
target_materials = material_endpoints.import_from_materialsproject(
    MATERIALS_PROJECT_API_KEY, TARGET_MATERIALS_PROJECT_IDS, owner_id
)


# ### Calculate Properties for "train materials"
#
# Create jobs for the "train materials".

# In[ ]:


compute = job_endpoints.get_compute(CLUSTER, PPN, NODES, QUEUE, TIME_LIMIT)
jobs = job_endpoints.create_by_ids(
    train_materials, band_gap_workflow_id, project_id, owner_id, JOB_NAME_PREFIX, compute
)


# Submit the jobs for execution.

# In[ ]:


for job in jobs:
    job_endpoints.submit(job["_id"])


# Monitor the jobs and print the status until they are all finished.

# In[ ]:


job_ids = [job["_id"] for job in jobs]
wait_for_jobs_to_finish(job_endpoints, job_ids)


# ### Build ML Train model
#
# Create ML Train job for the train materials.

# In[ ]:


name = "-".join((JOB_NAME_PREFIX, "train"))
material_ids = [m["_id"] for m in train_materials]
config = job_endpoints.get_config(material_ids, ml_train_workflow_id, project_id, owner_id, name, compute, True)
job = job_endpoints.create(config)


# Submit the train job for execution.

# In[ ]:


job_endpoints.submit(job["_id"])


# Monitor the job and print the status until it is done.

# In[ ]:


wait_for_jobs_to_finish(job_endpoints, [job["_id"]])


# ### Extract ML model as workflow
#
# The resulting trained model is extracted from the last unit (train with index 4) of the first job's subworkflow (ML: Train Model with index 0) and is further referred to as "ML predict workflow".

# In[ ]:


ml_predict_workflow = get_property_by_subworkflow_and_unit_indicies(
    property_endpoints, "workflow:ml_predict", job, 0, 4
)["data"]
ml_predict_workflow_id = ml_predict_workflow["_id"]


# Print ML predict workflow

# In[ ]:


display_JSON(ml_predict_workflow)


# ### Predict property using the model
#
# Create ML Predict job for the predict materials.

# In[ ]:


name = "-".join((JOB_NAME_PREFIX, "predict"))
material_ids = [m["_id"] for m in target_materials]
config = job_endpoints.get_config(material_ids, ml_predict_workflow_id, project_id, owner_id, name, compute, True)
job = job_endpoints.create(config)


# Submit the predict job for execution.

# In[ ]:


job_endpoints.submit(job["_id"])


# Monitor the job and print the status until its done.

# In[ ]:


wait_for_jobs_to_finish(job_endpoints, [job["_id"]])


# ### Extract predicted properties
#
# Predicted properties are extracted from the last unit (score with index 3) of the first job's subworkflow (ml_predict_subworkflow with index 0).

# In[ ]:


predicted_properties = get_property_by_subworkflow_and_unit_indicies(
    property_endpoints, "predicted_properties", job, 0, 3
)["data"]["values"]


# ### Flatten results
#
# The below for-loop iterates over the results and flatten them to form the final Pandas dataFrame.

# In[ ]:


table = []
for exabyte_id, properties in predicted_properties.items():
    material = next((m for m in target_materials if m["exabyteId"] == exabyte_id))
    band_gaps = next((v for v in properties if v["name"] == "band_gaps"))
    direct_gap = next((v for v in band_gaps["values"] if v["type"] == "direct"))["value"]
    indirect_gap = next((v for v in band_gaps["values"] if v["type"] == "indirect"))["value"]
    table.append(
        [material["_id"], material["name"], material["formula"], material["exabyteId"], direct_gap, indirect_gap]
    )


# ### Ouput results
#
# Create and print the final table as Pandas dataFrame.

# In[ ]:


headers = ["ID", "NAME", "FORMULA", "EXABYTE-ID", "DIRECT-GAP", "INDIRECT-GAP"]
df = pd.DataFrame(data=table, columns=headers)
html = dataframe_to_html(df)
html




#!/usr/bin/env python
# coding: utf-8

# # De Novo Generation of Bitter Molecules
# 
# In this webinar, we train a [Hidden Markov Model](https://en.wikipedia.org/wiki/Hidden_Markov_model) on a large corpus of SMILES strings, and use it to generate new SMILES strings. Finally, taking things one step further, we combine our newly-trained generative model with the bitterness classifier trained in the previous webinar. The end result of this generate->filter approach is a pipeline that generates new possibly-bitter molecules.

# In[]:


get_ipython().system('pip install optuna rdkit_pypi deepchem transformers matplotlib tqdm hmmlearn xgboost')


# In[]:


import os
from typing import List
import random
import urllib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import tqdm
import rdkit.Chem.Descriptors
from hmmlearn.hmm import MultinomialHMM
import xgboost

import rdkit
rdkit.RDLogger.DisableLog('rdApp.*')
from rdkit.Chem.Draw import IPythonConsole
from rdkit.Chem.Draw.MolDrawing import MolDrawing
from rdkit.Chem import Lipinski, Descriptors
import functools


# # Read in our corpus of SMILES strings
# 
# We have acquired a large datase of SMILES strings from the following repository: https://github.com/GLambard/Molecules_Dataset_Collection
# 
# We'll git clone, iterate over each file, and read in all of the SMILES strings we can find. In addition, for the purposes of keeping things simple in the webinar, we'll remove SMILES strings containing a colon (used for denoting the conjugated pi-bonds present in aromatic systems). This makes the problem a bit easier for our model to learn.

# In[]:


get_ipython().system('git clone https://github.com/Exabyte-io/Molecules_Dataset_Collection.git ../assets/Molecules_Dataset_Collection')


# In[]:


dataset_path = '../assets/Molecules_Dataset_Collection/latest'
files = [i for i in os.listdir(dataset_path) if i.endswith('csv')]

all_smiles = set()

for file in files:
    df = pd.read_csv(os.path.join(dataset_path,file))
    all_smiles.update(df.smiles.dropna())
    
all_smiles = set(smile for smile in all_smiles if ':' not in smile)
print(f"total of {len(all_smiles)} smiles")
print("Sample smiles:")
for count,smile in enumerate(all_smiles):
    print(smile)
    if count > 10:
        break


# # Import the SMILES Tokenizer from DeepChem
# 
# Next, we'll take advantage of the work of DeepChem, an open-source machine learning library focused on science. One of the tools that DeepChem provides is a SMILES tokenizer. This is used to convert a SMILES string into a list of individual pieces, or "tokens." For example, a single carbon atom may be a token, an opening parenthesis may be a token, etc. This is slightly different than simply separating by character, as tokens are allowed to contain multiple characters (e.g. \[Na+\] may be a token).

# In[]:


# Need to patch deepchem first per https://github.com/deepchem/deepchem/issues/2519
import re, importlib
deepchem_root_path=importlib.util.find_spec('deepchem').origin
smiles_tokenizer_path = os.path.join(os.path.dirname(deepchem_root_path), "feat", "smiles_tokenizer.py")
with open(smiles_tokenizer_path, "r") as inp:
    file_contents = inp.read()
with open(smiles_tokenizer_path, "w") as outp:
    for line in file_contents.split("\n"):
        outp.write(re.sub("max_len(\s|$)", "model_max_length", line) + "\n")
    
from deepchem.feat.smiles_tokenizer import SmilesTokenizer


# # Tokenize the Strings
# 
# The other important feature of the SMILES tokenizer is that it can also encode our tokenized SMILES strings. After we have tokenized our strings, we need to convert the individual tokens into numbers (for example, we may replace "C" with a "10" every time we see it, a "(" may become a "2", etc) for our ML algorithm to be able to use it. This process of taking a list of tokens, and mapping the strings to integers, is called "encoding."

# In[]:


class SmilesEncoder():
    def __init__(self):
        # Download Deepchem's vocabulary file if we don't already have it
        vocab_url = "https://raw.githubusercontent.com/deepchem/deepchem/master/deepchem/feat/tests/data/vocab.txt"
        self.vocab_path = "../assets/deepchem_smiles_vocab.txt"
        if not os.path.exists(self.vocab_path):
            with open(vocab_path, "wb") as outp:
                download = urllib.request.urlopen(vocab_url)
                outp.write(download.read())
                
                # Also, make sure colons are in our vocabulary
                outp.write(b"\n:")
                
        self.tokenizer = SmilesTokenizer(self.vocab_path)
        self.inversion_dict = {value:key for key,value in self.tokenizer.vocab.items()}
        
        self.ignored_range = list(range(self.tokenizer.vocab['[PAD]'], self.tokenizer.vocab['[MASK]']))
        
    def transform(self, X:str) -> List[int]:
        # Don't care about the start/end
        tokenized = self.tokenizer.encode(X)
        return tokenized
    
    def inverse_transform(self, X:List[int]) -> str:
        sep = self.tokenizer.vocab['[SEP]']
        if sep in X:
            end_index = np.random.choice(np.where(X==sep)[0])
            if end_index <= 1:
                end_index = -1
        else:
            end_index = -1
        result = "".join(self.inversion_dict[i] for i in X[1:end_index] if i not in self.ignored_range)
        return result   
enc = SmilesEncoder()


# In[]:


encoded = list(map(enc.transform, all_smiles))
np.random.shuffle(encoded)


# # Tune the Hidden Markov Model
# 
# As a first-order guess, we'll check a few number of hidden states for the Hidden Markov Model (HMM). We just do a gridsearch over models with 1-4 hidden states, and record the average log likelihood of the model when trained on a random subsample of the data. We train on a random subsample instead of the full dataset here primarily as a convenience for the purposes of demonstration, because the model may take a while to train as the dataset grows larger. To help get a better idea of the error, we repeat this process of randomly subsampling, training, and assessing with log-likelihood several times.

# In[]:


# Here, we're just going to use grid search, assesing the model accuracy via random subsampling from the dataset

def objective(n_states: int) -> int:
    """
    Optimize the number of hidden states. Train on a random subsample from the dataset, since this is rather expensive to run.
    """
    n_subsamples = 16
    subsample_size = 128
    
    total = 0
    for n in tqdm.trange(n_subsamples):
        model = MultinomialHMM(n_states, verbose=False, n_iter=64)
        x=np.concatenate([random.choice(encoded) for _ in range(subsample_size)]).reshape(-1,1)
        model.fit(x)
        total += model.score(x)
    mean = total / n_subsamples
    return mean

vals = []
n_max = 4
for n in range(1,n_max+1):
    print(f"{n} states", flush=True)
    vals.append(objective(n))


# In[]:


plt.plot(list(range(1,len(vals)+1)), vals)
plt.xlabel("Number of Hidden States")
plt.ylabel("Log Loss")
plt.show()
plt.close()


# # Train the Hidden Markov Model
# 
# Next, we'll take whichever number of hidden states made a model with the best log-likelihood. As we might expect, a more-complex HMM (more hidden states) captures the data a bit better - so we select 4 states.
# 
# Next, we'll train the model on the full dataset, until the error is either below 0.01 (the default in HMMlearn), or 1024 training steps have run.

# In[]:


best_n, best_loss = max(enumerate(vals, 1), key=lambda i: i[0])
print(f"Training with {best_n} hidden states")

model = MultinomialHMM(best_n, verbose=True, n_iter=1024)

# This is still a fairly expensive model to train, so we'll restrict ourselves to 10000 molecules in the dataset
model.fit(np.concatenate(encoded).reshape(-1,1))


# In[]:


# Now that we have the model, let's save it to disk (just in case)
import pickle
with open("../assets/smiles_markov_model.pkl", "wb") as outp:
    pickle.dump(model, outp)


# # Create a Molecule Pipeline
# 
# We'll then wrap our HMM into a generator function. This is an initial filter we place over the model, giving users the ability to filter generated models by the number of atoms.

# In[]:


def generate_molecules(min_heavy_atoms:int, max_heavy_atoms:int, num_mols:int, verbose:bool=True) -> List[rdkit.Chem.Mol]:
    """
    A pipeline that generates our molecules, from our generator!
    """
    pred_mols = []
    while len(pred_mols) < num_mols:
        sample, _ = model.sample(max_heavy_atoms)
        sample = sample.T[0]
        pred = enc.inverse_transform(sample)

        try:
            mol = rdkit.Chem.MolFromSmiles(pred)
        except:
            mol = None
        if mol and mol.GetNumAtoms() >= min_heavy_atoms:
            pred_mols.append(mol)
            if verbose:
                print(f"Found {len(pred_mols)} molecules. New molecule has weight {np.round((rdkit.Chem.Descriptors.MolWt(mol)),3)} amu")
    return pred_mols


# In[]:


molecules = generate_molecules(min_heavy_atoms=4, max_heavy_atoms=64, num_mols=8, verbose=True)
img = rdkit.Chem.Draw.MolsToGridImage(molecules, subImgSize=[256,256], molsPerRow=4)
img


# # Screen for Bitterness
# 
# Finally, we'll add one more filter to our model. Now that we have a set of knobs to turn for the molecule's complexity (number of atoms), we'll also add a filter for bitterness. Taking advantage of our trained model from the last webinar, we'll repeatedly generate models swith the generator, continuing until we have some user-defined set of potentially-bitter molecules.

# In[]:


with open("../assets/bittersweet_xgboost.pkl", "rb") as inp:
    pipeline = pickle.load(inp)
    
# Create a function to do the featurization and prediction
def predict_bitter_from_smiles(mol, model=pipeline):
    # The XGBoost classifier we trained in the last webinar takes in these features in this exact order
    descriptors = (Lipinski.NumAromaticRings,
                   Lipinski.NumHAcceptors,
                   Lipinski.NumHDonors,
                   Descriptors.MolWt,
                   Descriptors.MinPartialCharge,
                   Descriptors.MaxPartialCharge,
                   Descriptors.FpDensityMorgan1,
                   Descriptors.FpDensityMorgan2,
                   Descriptors.FpDensityMorgan3,
                  )
    
    features = [descriptor(mol) for descriptor in descriptors]
    result = model.predict([features])
    return result


# In[]:


bitter_candidates = []
n_desired_candidates = 4
while len(bitter_candidates) < n_desired_candidates:
    bitter_candidates += list(filter(predict_bitter_from_smiles, generate_molecules(min_heavy_atoms=4,
                                                          max_heavy_atoms=30,
                                                          num_mols=16,
                                                          verbose=False)
          ))
img = rdkit.Chem.Draw.MolsToGridImage(bitter_candidates[:n_desired_candidates], subImgSize=[256,256], molsPerRow=4)
img


# Of interest is that the pipeline generated a molecule that does indeed have a bitter taste. The first molecule in the above image is 2-propanol, which has a slightly bitter taste (per the MSDS as https://www2.atmos.umd.edu/~russ/MSDS/isopropanol_fisher.html). 



#
# This file contains common variables that are used inside the examples.
#
import json
import os

# General settings. For how the notebooks operate.

# use_interactive_JSON_viewer: Whether to use the IPython interactive viewer, or print in plaintext.

use_interactive_JSON_viewer = True

# Account settings. Need a one-time adjustment for examples to work. These should be set in settings.json

# ACCOUNT_ID: Account ID. See get_authentication_params.ipynb example for more information.
# AUTH_TOKEN: Account authentication token. See get_authentication_params.ipynb for more information.
# MATERIALS_PROJECT_API_KEY: Materials project API key.
# See https://legacy.materialsproject.org/open for more information.

# Load variables from the settings.json file

absolute_path_to_settings_json_file = os.path.join(os.path.dirname(__file__), "settings.json")
assert absolute_path_to_settings_json_file
with open(absolute_path_to_settings_json_file) as settings_json_file:
    settings_json_config = json.load(settings_json_file)

# The variables below are defined in the first code cell
# in either the Google Colab or Jupyter notebook.
#
# In Google Colab, these variables must be filled out in the 'Authorization Form' section of the notebook,
# which are added to the environment variables.
#
# If using Jupyter, these variables can be left to their default values in the code cell, but the user
# should change these values in the settings.json file located in the examples folder.
#
# We prioritize the environment variables for Google Colab, and fall back to the settings from JSON for Jupyter.

ACCOUNT_ID = os.getenv("ACCOUNT_ID", settings_json_config.get("ACCOUNT_ID"))
AUTH_TOKEN = os.getenv("AUTH_TOKEN", settings_json_config.get("AUTH_TOKEN"))
MATERIALS_PROJECT_API_KEY = os.getenv(
    "MATERIALS_PROJECT_API_KEY", settings_json_config.get("MATERIALS_PROJECT_API_KEY")
)
ORGANIZATION_ID = os.getenv("ORGANIZATION_ID", settings_json_config.get("ORGANIZATION_ID"))

# Advanced settings. Should not need adjustments.

# HOST: Hostname of the RESTful API server. Defaults to platform.mat3ra.com.
# PORT: The port RESTful API server is listening on. Defaults to 443.
# VERSION: RESTFul API version. Defaults to 2018-10-01.
# SECURE: Whether to use secure connection. Defaults to True.

PORT = os.environ.get("API_PORT", 443)
SECURE = os.environ.get("API_SECURE") != "false"
VERSION = os.environ.get("API_VERSION", "2018-10-01")
HOST = os.environ.get("API_HOST", "platform.mat3ra.com")
ENDPOINT_ARGS = [HOST, PORT, ACCOUNT_ID, AUTH_TOKEN, VERSION, SECURE]



"""TITLE: Import of all supported packages."""
"""BLOCK: Install Packages"""
"""
This scripts installs the required packages for a basic Pyodide environment for Materials structure manipulation.
It uses micropip - a Python package manager for Pyodide.
"""
import micropip

packages = [
    "https://files.mat3ra.com:44318/uploads/pymatgen-2023.9.10-py3-none-any.whl",
    "https://files.mat3ra.com:44318/web/pyodide/spglib-2.0.2-py3-none-any.whl",
    "https://files.pythonhosted.org/packages/d9/0e/2a05efa11ea33513fbdf4a2e2576fe94fd8fa5ad226dbb9c660886390974/ruamel.yaml-0.17.32-py3-none-any.whl",
    "ase==3.22.1",
    "networkx==3.2.1",
    "monty==2023.11.3",
    "scipy==1.11.1",
    "lzma",
    "tabulate==0.9.0",
    "sqlite3",
    "sympy==1.12",
]


async def install_package(pkg):
    """
    Installs a package in a Pyodide environment.
    Args:
        pkg: The name of the package to install.

    Returns:
        None
    """
    is_url = pkg.startswith("http://") or pkg.startswith("https://")
    are_dependencies_installed = not is_url
    await micropip.install(pkg, deps=are_dependencies_installed)
    # Extract package name for printing
    pkg_name = pkg.split("/")[-1].split("-")[0] if is_url else pkg.split("==")[0]
    print(f"Installed {pkg_name}")


for package in packages:
    await install_package(package)



#!/usr/bin/env python
# coding: utf-8

# # Predict Adsorption Energies with Machine Learning
# 
# This notebook was created to facilitate our April 2021 webinar, which detailed the creation of a simple model of CO adsorption. The video can be found in the following link:
# 
# https://www.youtube.com/watch?v=wOrqnR8UAbQ
# 
# The descriptors and dataset used in the machine learning portion of this notebook are based on the following work:
# 
# Dean, J.; Taylor, M. G.; Mpourmpakis, G. Unfolding Adsorption on Metal Nanoparticles: Connecting Stability with catalysis. Sci Adv 2019, 5 (9), eaax5101. DOI: doi.org/10.1126/sciadv.aax5101
# 
# In this notebook example, we automate the placement of CO molecules to several Cu slabs. We then submit all of these adsorbed states to the cluster as DFT geometry optimization calculations using VASP.

# In[]:


import pandas
import numpy as np
import sklearn.preprocessing, sklearn.model_selection, sklearn.linear_model
import matplotlib.pyplot as plt

import sklearn.neural_network
import sklearn.linear_model
import sklearn.kernel_ridge 


# # Load the Data
# 
# We begin by loading the data from the CSV using Pandas, and split it into the target and descriptors.

# In[]:


datafile = "data_to_train_with.csv"
target_column_name = "PBE_BE_eV"

data = pandas.read_csv(datafile)
target = data.pop(target_column_name).to_numpy()
target = target.reshape(-1, 1)  # Reshape array from a row vector into a column vector
descriptors = data.to_numpy()


# # Train/Test Split
# 
# Next, we perform an 80/20 train/test split on the dataset. 80% is kept for training, and 20% is kept for prediction.

# In[]:


percent_held_as_test = 0.2
train_descriptors, test_descriptors, train_target, test_target = sklearn.model_selection.train_test_split(descriptors,
                                                                                                          target,
                                                                                                          test_size=percent_held_as_test)


# # Standardize the Data
# 
# Next, we will scale the data such that it has a mean of 0 and a standard deviation of 1. We will also save the scaler, so that we can unscale our data at the end of the pipeline (for parity plot generation).

# In[]:


scaler = sklearn.preprocessing.StandardScaler
descriptor_scaler = scaler()
train_descriptors = descriptor_scaler.fit_transform(train_descriptors)
test_descriptors = descriptor_scaler.transform(test_descriptors)

target_scaler = scaler()
train_target = target_scaler.fit_transform(train_target)
test_target = target_scaler.transform(test_target)


# # Model Selection and Regression
# 
# Next, to demonstrate the programmatic selection of models, we will choose between three types of model:
# - Ridge Regression
# - Kernelized Ridge Regression
# - Neural Network

# In[]:



train_target = train_target.flatten()
test_target = test_target.flatten()

# Initialize the model
models = {"Neural Network" : sklearn.neural_network.MLPRegressor(hidden_layer_sizes=(100,)),
          "Ridge Regression" : sklearn.linear_model.Ridge(alpha=1.0),
          "Kernelized Ridge Regression": sklearn.kernel_ridge.KernelRidge(alpha=1.0,
                                                                          kernel="linear")}
best_name = None
best_model = None
best_rmse = np.inf
for name, model in models.items():
    # Train the model and save
    model.fit(train_descriptors, train_target)
    train_predictions = model.predict(train_descriptors)
    test_predictions = model.predict(test_descriptors)

    # Scale predictions so they have the same shape as the saved target
    train_predictions = train_predictions.reshape(-1, 1)
    test_predictions = test_predictions.reshape(-1, 1)

    test_target = test_target.reshape(-1, 1)
    y_true = target_scaler.inverse_transform(test_target)
    y_pred = target_scaler.inverse_transform(test_predictions)

    # RMSE
    mse = sklearn.metrics.mean_squared_error(y_true, y_pred)
    rmse = np.sqrt(mse)
    print(f"{name}: RMSE = {rmse}")
    
    if rmse < best_rmse:
        best_name = name
        best_model = model
        best_rmse = rmse
        
print("Best Model: ", best_name)

model = best_model


# # Parity plot                               
#                                                                    
# Finally, we generate a parity plot of predicted versus actual values for the training and testing set.

# In[]:


# Un-transform the data
train_target = target_scaler.inverse_transform(train_target)
train_predictions = target_scaler.inverse_transform(train_predictions)
test_target = target_scaler.inverse_transform(test_target)
test_predictions = target_scaler.inverse_transform(test_predictions)

# Plot the data
plt.scatter(train_target, train_predictions, c="#203d78", label="Training Set")
plt.scatter(test_target, test_predictions, c="#67ac5b", label="Testing Set")
plt.xlabel("Actual Value")
plt.ylabel("Predicted Value")

# Scale the plot
target_range = (min(min(train_target), min(test_target)),
                max(max(train_target), max(test_target)))
predictions_range = (min(min(train_predictions), min(test_predictions)),
                     max(max(train_predictions), max(test_predictions)))

limits = (min(min(target_range), min(target_range)),
          max(max(predictions_range), max(predictions_range)))
plt.xlim = (limits[0], limits[1])
plt.ylim = (limits[0], limits[1])

# Draw a parity line, as a guide to the eye
plt.plot((limits[0], limits[1]), (limits[0], limits[1]), c="black", linestyle="dotted", label="Parity")
plt.legend()

# Save the figure
plt.show()




#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/workflow/qe_scf_calculation.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# 

# # Quantum Espresso SCF calculation via API
# 

# # Complete Authorization Form and Initialize Settings
# 
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
# 
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
# 
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
# 
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
# 
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
# 
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# In[ ]:


from utils.settings import ENDPOINT_ARGS, ACCOUNT_ID
from utils.generic import wait_for_jobs_to_finish

from exabyte_api_client.endpoints.workflows import WorkflowEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.jobs import JobEndpoints


# In[ ]:


# Initialize a helper class to interact with WorkflowEndpoints
workflow_endpoints = WorkflowEndpoints(*ENDPOINT_ARGS)
material_endpoints = MaterialEndpoints(*ENDPOINT_ARGS)
job_endpoints = JobEndpoints(*ENDPOINT_ARGS)


# #### Create a Quantum Espresso workflow for SCF calculation
# 
# Below we provide our Quantum Espresso input file via a bash script.
# 
# Note that we provide the pseudo potential file via a downloadable url. We are working on supporting direct uploading of pseudo potential file from local file system, but it is currently not available. Below are some of the sources to find pseudo potentials:
# 
# - [Pseudopotentials library at Quantum Espresso](https://www.quantum-espresso.org/pseudopotentials/)
# - [Standard solid-state pseudopotentials (SSSP) library](https://www.materialscloud.org/discover/sssp/table/efficiency)
# - [Pseudo dojo library](http://www.pseudo-dojo.org)
# - [GBRV pseudo potential library](https://www.physics.rutgers.edu/gbrv/)

# In[ ]:


# user modifiable part
script = r"""#!/bin/bash

# switch to the job working directory
cd $PBS_O_WORKDIR

# ----------------------- QUANTUM ESPRESSO INPUT FILE ------------------------ #
cat > pw.in << EOF
&CONTROL
  calculation = 'scf',
  prefix = 'silicon',
  outdir = './tmp/'
  pseudo_dir = './'
/

&SYSTEM
  ibrav =  2,
  celldm(1) = 10.26,
  nat =  2,
  ntyp = 1,
  ecutwfc = 30
  nbnd = 8
/

&ELECTRONS
  conv_thr = 1e-8,
  mixing_beta = 0.6
/

ATOMIC_SPECIES
  Si 28.086 Si.pz-vbc.UPF

ATOMIC_POSITIONS (alat)
  Si 0.0 0.0 0.0
  Si 0.25 0.25 0.25

K_POINTS (automatic)
  7 7 7 0 0 0
EOF
# -------------------------- PSEUDO POTENTIAL FILE --------------------------- #
# provide a downloadable link for the pseudo potential file
wget https://media.githubusercontent.com/media/exabyte-io/api-examples/dev/examples/assets/Si.pz-vbc.UPF
# --------------------------------- RUN JOB ---------------------------------- #
# load required module
module add espresso/63-i-174-impi-044

mpirun -np $PBS_NP pw.x -in pw.in | tee pw.out
"""


# In[ ]:


# populate workflow from a template
import json

with open("../assets/bash_workflow_template.json", "r") as f:
    WORKFLOW_BODY = json.load(f)

WORKFLOW_BODY["subworkflows"][0]["units"][0]["input"][0]["content"] = script


# In[ ]:


# create workflow
WORKFLOW_RESP = workflow_endpoints.create(WORKFLOW_BODY)


# ### Create and submit job
# 
# Below user can specify the project name and compute parameters such as `queue`, number of `nodes` and number of processors `ppn` per node. Find more about compute parameters [here](https://docs.mat3ra.com/infrastructure/compute/parameters/).

# In[ ]:


# job creation payload
JOB_BODY = {
    "name": "SCF Calculation",
    "compute": {"ppn": 4, "nodes": 1, "queue": "OR"},
    "_project": {"slug": "pranab-default"},
    "workflow": WORKFLOW_RESP["_id"],
}


# In[ ]:


# create job
JOB_RESP = job_endpoints.create(JOB_BODY)


# In[ ]:


# TODO: NOT SUPPORTED YET, upload pseudo potential file
# with open('../assets/Si.pz-vbc.UPF', 'r') as f:
#     PP_FILE = f.read()
#     data = json.dumps({"files": PP_FILE})
#     FILE_RESP=job_endpoints.insert_output_files(JOB_RESP["_id"], data)


# In[ ]:


# submit job to run
job_endpoints.submit(JOB_RESP["_id"])


# In[ ]:


# monitor job and wait for it to be finished
wait_for_jobs_to_finish(job_endpoints, [JOB_RESP["_id"]], 30)


# #### Get output file, perform post processing, and make plots

# In[ ]:


files = job_endpoints.list_files(JOB_RESP["_id"])
for file in files:
    if file["name"] == "pw.out":
        output_file_metadata = file

import urllib

server_response = urllib.request.urlopen(output_file_metadata["signedUrl"])
output_file_bytes = server_response.read()
output_file = output_file_bytes.decode(encoding="UTF-8")


# In[ ]:


energy = []
len_energy = len("total energy")
for line in output_file.split("\n"):
    if line.strip().lstrip("!")[:len_energy] == "total energy":
        energy.append(float(line.split("=")[1].rstrip("Ry")))


# In[ ]:


# plot energy with iteration step
import matplotlib.pyplot as plt

get_ipython().run_line_magic('matplotlib', 'inline')

plt.plot(energy)
plt.xlabel("Number of iteration")
plt.ylabel("Energy (Ry)")
plt.show()




import os
import re
from urllib.parse import unquote, urlparse

import requests


def get_notebook_info() -> dict:
    """
    Get the information about a currently running notebook in Google Colab.
    Args:
        None
    Return:
        a dict with notebook info.
    """
    # ip = socket.gethostbyname(socket.gethostname())  # 172.28.0.12
    ip = os.getenv("COLAB_JUPYTER_IP")
    response = requests.get(f"http://{ip}:9000/api/sessions").json()[0]

    notebook_name = response["name"]
    path = urlparse(unquote(response["path"]).split("=")[1]).path
    parsed = re.findall("(.*)/blob/(.*)/examples/(.*)", path)[0]
    github_org_repo = parsed[0][1:]  # remove leading /
    branch_name = parsed[1]
    notebook_path = f"examples/{parsed[2]}"

    return dict(
        notebook_name=notebook_name,
        notebook_path=notebook_path,
        branch_name=branch_name,
        github_org_repo=github_org_repo,
    )


def print_notebook_path() -> None:
    """
    A proxy function used for a single string return when
    the corresponding entry-point script in 'setup.py' is called.

    Args:
        None
    Return:
        None
    """
    # 'return get_notebook_info()["notebook_path"]' returns a non-zero exit code; using 'print' instead.
    print(get_notebook_info()["notebook_path"])



# This module defines a set of common functions which are used in other examples.
import datetime
import json
import os
import time
import urllib.request
import uuid
from typing import List, Union

from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.properties import PropertiesEndpoints
from IPython.display import HTML, display
from pandas import DataFrame
from pandas.io.formats.style import Styler
from tabulate import tabulate

from . import settings

# GENERIC UTILITIES


def update_json_file_kwargs(path_to_json_file: str = "settings.json", **kwargs) -> None:
    """
    This function updates settings.json for a given kwargs if kwargs
    contains variables different from those already in json

    Args:
        path_to_json_file (str): the path to the json file to be updated
        **kwargs (dict): A dict of keyword arguments

    Returns:
        None
    """

    # 1. Assert the json file is where we think it is
    assert os.path.isfile(path_to_json_file)

    # 2. Load settings.json
    with open(path_to_json_file) as settings_json_file:
        variables = json.load(settings_json_file)

    # 3. Update json file if kwargs contains new variables
    if kwargs != variables:
        updated_variables = {**variables, **kwargs}
        with open(path_to_json_file, "w") as settings_json_file:
            json.dump(updated_variables, settings_json_file, indent=4)


def save_files(job_id: str, job_endpoint: JobEndpoints, filename_on_cloud: str, filename_on_disk: str) -> None:
    """
    Saves a file to disk, overwriting any files with the same name as filename_on_disk

    Args:
        job_id (str): ID of the job
        job_endpoint (JobEndpoints): Job endpoint object from the Exabyte API Client
        filename_on_cloud (str): Name of the file on the server
        filename_on_disk (str): Name the file will be saved to

    Returns:
        None
    """
    files = job_endpoint.list_files(job_id)
    for file in files:
        if file["name"] == filename_on_cloud:
            file_metadata = file

    # Get a download URL for the CONTCAR
    signed_url = file_metadata["signedUrl"]

    # Download the contcar to memory
    server_response = urllib.request.urlopen(signed_url)

    # Write it to disk
    with open(filename_on_disk, "wb") as outp:
        outp.write(server_response.read())


# JOB UTILITIES


def get_jobs_statuses_by_ids(endpoint: JobEndpoints, job_ids: List[str]) -> List[str]:
    """
    Gets jobs statues by their IDs.

    Args:
        job_endpoint (JobEndpoints): Job endpoint object from the Exabyte API Client
        job_ids (list): list of job IDs to get the status for

    Returns:
        list: list of job statuses
    """
    jobs = endpoint.list({"_id": {"$in": job_ids}}, {"fields": {"status": 1}})
    return [job["status"] for job in jobs]


def wait_for_jobs_to_finish(endpoint: JobEndpoints, job_ids: list, poll_interval: int = 10) -> None:
    """
    Waits for jobs to finish and prints their statuses.
    A job is considered finished if it is not in "pre-submission", "submitted", or, "active" status.

    Args:
        job_endpoint (JobEndpoints): Job endpoint object from the Exabyte API Client
        job_ids (list): list of job IDs to wait for
        poll_interval (int): poll interval for job information in seconds. Defaults to 10.
    """
    print("Wait for jobs to finish, poll interval: {0} sec".format(poll_interval))
    while True:
        statuses = get_jobs_statuses_by_ids(endpoint, job_ids)

        errored_jobs = len([status for status in statuses if status == "error"])
        active_jobs = len([status for status in statuses if status == "active"])
        finished_jobs = len([status for status in statuses if status == "finished"])
        submitted_jobs = len([status for status in statuses if status == "submitted"])

        headers = ["TIME", "SUBMITTED-JOBS", "ACTIVE-JOBS", "FINISHED-JOBS", "ERRORED-JOBS"]
        now = datetime.datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
        row = [now, submitted_jobs, active_jobs, finished_jobs, errored_jobs]
        print(tabulate([row], headers, tablefmt="grid", stralign="center"))

        if all([status not in ["pre-submission", "submitted", "active"] for status in statuses]):
            break
        time.sleep(poll_interval)


# WORKFLOW


def copy_bank_workflow_by_system_name(endpoint: BankWorkflowEndpoints, system_name: str, account_id: str) -> dict:
    """
    Copies a bank workflow with given ID into the account's workflows.

    Args:
        endpoint (endpoints.bank_workflows.BankWorkflowEndpoints): an instance of BankWorkflowEndpoints class
        system_name (str): workflow system name.
        account_id (str): ID of account to copy the bank workflow into.

    Returns:
        dict: new account's workflow
    """
    bank_workflow_id = endpoint.list({"systemName": system_name})[0]["_id"]
    return endpoint.copy(bank_workflow_id, account_id)["_id"]


# PROPERTY


def get_property_by_subworkflow_and_unit_indicies(
    endpoint: PropertiesEndpoints, property_name: str, job: dict, subworkflow_index: int, unit_index: int
) -> dict:
    """
    Returns the property extracted in the given unit of the job's subworkflow.

    Args:
        endpoint (endpoints.properties.PropertiesEndpoints): an instance of PropertiesEndpoints class.
        property_name (str): name of property to extract.
        job (dict): job config to extract the property from.
        subworkflow_index (int): index of subworkflow to extract the property from.
        unit_index (int): index of unit to extract the property from.

    Returns:
        dict: extracted property
    """
    unit_flowchart_id = job["workflow"]["subworkflows"][subworkflow_index]["units"][unit_index]["flowchartId"]
    return endpoint.get_property(job["_id"], unit_flowchart_id, property_name)


# DISPLAY UTILITIES


def dataframe_to_html(df: DataFrame, text_align: str = "center") -> Styler:
    """
    Converts Pandas dataframe to HTML.
    See https://pandas.pydata.org/pandas-docs/stable/style.html for more information about styling.

    Args:
        df (pd.DataFrame): Pandas dataframe.
        text_align (str): text align. Defaults to center.
    """
    styles = [
        dict(selector="th", props=[("text-align", text_align)]),
        dict(selector="td", props=[("text-align", text_align)]),
    ]
    return df.style.set_table_styles(styles)


def display_JSON(
    obj: Union[dict, list], interactive_viewer: bool = settings.use_interactive_JSON_viewer, level: int = 2
) -> None:
    """
    Displays JSON, either interactively or via a text dump to Stdout.

    The interactive viewer is based on https://github.com/mljar/mercury/blob/main/mercury/widgets/json.py.

    Args:
        obj (dict): Object to display as nicely-formatted JSON
        interactive (bool): Whether to use the interactive viewer or not
    """
    if interactive_viewer:
        if isinstance(obj, (dict, list)):
            json_str = json.dumps(obj)
        else:
            json_str = obj

        id = str(uuid.uuid4())

        web_dir = os.path.join(os.path.dirname(__file__), "web")

        with open(os.path.join(web_dir, "renderjson.css")) as fp:
            css = fp.read()

        with open(os.path.join(web_dir, "renderjson.js")) as fp:
            js = fp.read()

        display(HTML(f'<style>{css}</style><div id="{id}"></div>'))
        display(
            HTML(
                f"<script>{js} "
                f"renderjson.set_show_to_level({str(level)}); "
                f'renderjson.set_icons("▸","▾"); '
                f'document.getElementById("{id}").appendChild(renderjson({json_str}))</script>'
            )
        )
    else:
        print(json.dumps(obj, indent=4))



#!/usr/bin/env python
# coding: utf-8

# # Predicting Perovskite Volumes
# 
# This dataset comes from a perovskite database hosted by the NOMAD lab (https://nomad-lab.eu/). In this notebook, we'll featurize a set of crystal cells automatically using XenonPy, and we'll create a model to predict the volume automatically using TPOT.
# 
# Let's start off by installing nomad-lab from pip. We'll also need to install TPOT and XenonPy

# In[]:


get_ipython().system('pip install nomad-lab tpot xenonpy')


# # Import Packages
# We'll import some standard Python ML packages.
# 
# In addition, we'll import a couple of others:
# - XenonPy (https://xenonpy.readthedocs.io/en/latest/index.html) is a materials science toolkit similar to RDKit, generating features for materials inputs
# - TPOT (http://epistasislab.github.io/tpot/) is an AutoML tool. Using a genetic algorithm, it acts as a computerized data-scientist, automatically testing and selecting a model for a dataset.
# - Nomad API (https://nomad-lab.eu/prod/rae/docs/index.html) allows access to NOMAD's large database of DFT calculations
# 
# We'll also specify a random seed, for reproducibility.

# In[]:


import re 

import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import tpot
import sklearn.pipeline

import ase
import ase.io
import xenonpy.datatools
from xenonpy.descriptor import Compositions
from nomad.client import ArchiveQuery

RANDOM_SEED = 42


# # Query NOMAD
# 
# We'll first query NOMAD for a database named "stable inorganic perovskites," which has 144 materials inside of it.
# 
# Then, we'll convert each feature to an ASE atoms object, for use in featurization down the road.
# 
# Finally, we'll save it all to a Pandas dataframe.

# In[]:


# Query NOMAD for a perovskites dataset

search = "stable inorganic perovskites"
query = ArchiveQuery(
    query = {
        "datasets": search
    },
    per_page = 144,
    max = 144
)

# Convert the query to ASE atoms objects
def to_atoms(result):
    """
    Function that takes in a result from NOMAD, and turns it into an ASE atom object.
    The object is based on the standardized unit cell reported by NOMAD.
    
    Args:
        result: The result of a query
    
    Returns:
        The result as an ASE Atoms object
    """
    # Get the description of the standardized unit cell
    system = result.section_run[-1].section_system[-1].section_symmetry[-1].section_std_system[-1]  
    
    # Extract structural parameters
    unit_cell = system.lattice_vectors_std
    fractional_coordinates = system.atom_positions_std
    atomic_symbols = [ase.data.chemical_symbols[atomic_number] for atomic_number in system.atomic_numbers_std]
    
    # Create and return the atoms object
    atoms = ase.Atoms(symbols=atomic_symbols, scaled_positions=fractional_coordinates, cell=unit_cell, pbc=[True, True, True])
    return atoms
    
atoms = [to_atoms(result) for result in query]

raw_data = {
    "Formula": [],
    "Atoms_Object": [],
    "Volume": []
}
for crystal in atoms:
    raw_data["Formula"].append(crystal.get_chemical_formula())
    raw_data["Atoms_Object"].append(crystal)
    raw_data["Volume"].append(crystal.cell.volume)
raw_data = pd.DataFrame.from_dict(raw_data)

# Normalize volume to empirical formula unit. Divide by the number of atoms and multiply by 5 (because these are all ABO3)
raw_data.Volume = 5 * raw_data.Volume / raw_data.Atoms_Object.apply(lambda atoms: len(atoms))
raw_data.head()


# # Feature Engineering
# 
# Next, we'll featurize our dataset using XenonPy. In a real-world scenario, this is where we should be spending much of our time. Good features lead to good models.
# 
# In this case, we're going to just use XenonPy to generate 290 features per entry. In a nutshell, it takes in the composition of each formula unit, and then looks up many descriptors (such as ionization energies).
# Then, it derives various properties, such as the minimum, maximum, average, etc. For example, the average 

# In[]:


# Featurize with XenonPy
# Xenonpy needs to download datasets for its featurizers to work
xenonpy.datatools.preset.sync("elements_completed")
element_cal = Compositions()

def featurize(atoms: ase.Atoms):
    # Get the chemical formula as a dict
    symbols = atoms.get_chemical_symbols()
    counts = {symbol: symbols.count(symbol) for symbol in symbols}
    
    # Do the featurization
    result = element_cal.transform([counts])
    return result

new_results = pd.DataFrame()
for index, atoms in zip(raw_data.index, raw_data.Atoms_Object):
    results = featurize(atoms)
    results.index = [index]
    new_results = new_results.append(results)

data = raw_data[["Volume"]].join(new_results)
data.to_csv("perovskite_data.csv")
data.head()


# # Correlation Chart
# 
# Below we find a correlation chart, giving a high-level overview of some of the correlations found in the ata. Because there are so many features, it's difficult to determine the individual correlations. That said, what we can see is that we're in a highly-correlated feature space - this can easily be observed by noting the relatively large area consisting of bright yellow and dark blue colors in the plot.

# In[]:


# Let's plot the correlations
matplotlib.rcParams['figure.dpi'] = 300
matplotlib.rcParams['axes.grid'] = False
cax = plt.matshow(data.corr(), cmap="plasma")
plt.gca().set_xticklabels(data.columns, fontsize=1)
plt.setp(plt.gca().get_xticklabels(), rotation=90)
plt.gca().set_yticklabels(data.columns, fontsize=1)
plt.xticks(range(len(data.columns)))
plt.yticks(range(len(data.columns)))
plt.colorbar(label="Correlation")
plt.show()


# # Train/Test Split
# 
# Next, we'll do a train/test split of the dataset. We don't have a lot of data, but we'll try a 50/50 train/test split. We're using a lot of features together with a highly flexible model (TPOT can choose between many different models and many different hyperparameters), so we want it to be very apparent if we're at all overfitting.

# In[]:


# Train/Test Split
train, test = sklearn.model_selection.train_test_split(data, test_size=0.5, random_state=RANDOM_SEED)

x_cols = data.drop(columns=["Volume"]).columns
y = train.Volume.to_numpy()
x = train[x_cols].to_numpy()


# # TPOT Regression
# 
# Next, we'll use TPOT, a genetic algorithm that automatically selects models and their hyperparameters. We'll use the following settings:
# - `generations = None`, to place no limit on the number of possible generations
# - `population_size = 20`, a small population size for exploratory purposes
# - `max_eval_time_mins = 100/60/1000`, again a short evaluation time of 10ms, just for exploring this technique on a laptop
# - `cv=len(x)`, to use leave-one-out cross-validation (LOOCV). It generally performs well on small datasets like ours, compared to 5/10-fold CV
# - `scoring` is set to the negative root mean squared error of the model, which will give us the (negative) error of the model we might expect, in Å^3
# - `n_jobs=-1`, to use as many jobs as there are cores on the machine
# 
# Finally, we'll set `max_time_mins` to 2, to run the pipeline for a grand total of 2 minutes.

# In[]:


model = tpot.TPOTRegressor(
    generations=None,
    population_size=100,
    max_eval_time_mins = 100/60/1000,
    max_time_mins=5,
    cv=len(x), # LOOCV
    verbosity=2,
    scoring="neg_root_mean_squared_error",
    config_dict=tpot.config.regressor_config_dict,
    n_jobs=-1,
    random_state=RANDOM_SEED
)
model.fit(x, y)


# # Training is Done!
# 
# Now, we'll print the best pipeline that was found. At the time this notebook was created (we used a random seed of 42 for the train/test split and the pipeline, so this should be reproducible), we find the following steps:
# - Feature selection based on the Family-Wise Error rate. Specifically, Sklearn's implementation uses the [Bonferroni Criterion](https://en.wikipedia.org/wiki/Bonferroni_correction). Given a significance alpha, a feature is kept if its p-value (w.r.t. the target varible) is below alpha divided by the number of samples in the training set. In this case, alpha is chosen to be 0.02.
# - Scaling each feature to be between 0 and 1, a fairly standard choice in ML work
# - Linear Support Vector Regression
# 
# Overall, this is a very reasonable result for this type of regression problem, one that a human would believably choose to evaluate when looking at different potential models. We see the elements of a basic ML pipeline here: feature selection, followed by some scaling of the data, followed by a model.
# 
# Moreover, the pipeline considers a fairly conservative feature selection approach, with a more-stringengent signifiance than normal (a=0.02 versus a=0.05), which will help in this situation, where we have many different features compared to the number of training samples.
# 
# The chosen model is also reasonable in this case, as the L2 loss function will help with the highly-dimensional space that we're fitting.

# In[]:


print("Best Pipeline")
print(model.fitted_pipeline_)

selected_features = data.drop(columns="Volume").columns[model.fitted_pipeline_[0].get_support()]

print(f"Model selected {len(selected_features)} features")
print(selected_features)


# # Draw a Parity Plot
# 
# Finally, we'll evaluate our model against the test set. Recall that we left out 50% of our testing set, because we want some extra insurance (beyond the LOOCV used during training) against overfitting. We can see in the below plot that in general, we do pretty well.
# 
# In fact, our test-set mean-absolute-percentage-error is 3.9%. This means that on average, we're off by 3.9% with respect to the testing set.

# In[]:


# Print some model statistics
train_preds = model.predict(x)
test_preds = model.predict(test[x_cols].to_numpy())
test_mape = sklearn.metrics.mean_absolute_percentage_error(y_true=test.Volume, y_pred=test_preds)
print(f"Test MAPE = {test_mape}")

# Parity Plot
plt.scatter(train_preds, y, label="Training Set")
plt.scatter(test_preds, test.Volume, label="Test Set")

# Ensure plot is square
lims = [0,
        max(max(y), max(test.Volume.to_numpy()))]
plt.xlim(lims)
plt.ylim(lims)

# Parity Line
plt.plot([lims[0], lims[1]], [lims[0], lims[1]],
         c="black", linestyle="dashed", label="Parity")

# Misc Drawing
plt.legend()
plt.xlabel("Predicted Volume (Å^3) / Formula Unit")
plt.ylabel("Actual Volume (Å^3) / Formula Unit")
plt.show() #Make these larger


# # Make a Prediction!
# 
# Generally, we train models to do predictions later. To test this out, we'll investigate Samarium Nickle Oxide. According to Materials Project ( https://materialsproject.org/materials/mp-22495/# ), this material has a volume of 226.362 Å^3. Normalized by the number of formula units, the volume is 60.105 Å^3 per formula unit.
# 
# We find that our volume is predicted to be 59.59 Å^3 per formula unit.

# In[]:


comp = {
    "Sm": 4,
    "Ni": 4,
    "O": 12,
}

generated_features = element_cal.transform([comp])

# Print the predicted volume
predicted_volume = model.predict(generated_features.to_numpy())[0]
known_volume = 226.362
known_volume_scaled = 5 * known_volume / (4 + 4 + 12)
print(f"Predicted Volume is: {round(predicted_volume,3)} Å^3 / formula unit")
print(f"Known Volume is {round(known_volume_scaled,3)} Å^3 / formula unit")

# Print the features we generated  
generated_features.to_csv("for_prediction_on_cluster.csv")
generated_features


# In[ ]:







import json
import os
from enum import Enum
from typing import Any, Dict, Optional

from IPython.display import Javascript, display

UPLOADS_FOLDER = "uploads"


class EnvironmentEnum(Enum):
    PYODIDE = "pyodide"
    PYTHON = "python"


# Environment detection
# default value for env.HOME from https://pyodide.org/en/stable/usage/api/js-api.html
ENVIRONMENT = EnvironmentEnum.PYODIDE if os.environ.get("HOME") == "/home/pyodide" else EnvironmentEnum.PYTHON

if ENVIRONMENT == EnvironmentEnum.PYODIDE:
    import micropip

if ENVIRONMENT == EnvironmentEnum.PYTHON:
    import subprocess
    import sys


async def install_package_pyodide(pkg: str, verbose=True):
    """
    Install a package in a Pyodide environment.
    Args:
        pkg (string): The name of the package to install.
        verbose (bool): Whether to print the name of the installed package.
    """
    is_url = pkg.startswith("http://") or pkg.startswith("https://") or pkg.startswith("emfs:/")
    are_dependencies_installed = not is_url
    await micropip.install(pkg, deps=are_dependencies_installed)
    pkg_name = pkg.split("/")[-1].split("-")[0] if is_url else pkg.split("==")[0]
    if verbose:
        print(f"Installed {pkg_name}")


def install_package_python(pkg: str, verbose=True):
    """
    Install a package in a standard Python environment.
    Args:
        pkg (string): The name of the package to install.
        verbose (bool): Whether to print the name of the installed package.
    """
    subprocess.check_call([sys.executable, "-m", "pip", "install", pkg])
    if verbose:
        print(f"Installed {pkg}")


async def install_packages(notebook_name: str, requirements_path="config.yml", verbose=True):
    """
    Install the packages listed in the requirements file for the notebook with the given name.
    Args:
        notebook_name (string): The name of the notebook for which to install packages.
        requirements_path (string): The path to the requirements file.
        verbose (bool): Whether to print the names of the installed packages and status of installation.
    """
    if ENVIRONMENT == EnvironmentEnum.PYODIDE:
        await micropip.install("pyyaml")
        # PyYAML has to be installed before being imported in Pyodide and can't appear at the top of the file
    import yaml

    base_path = os.getcwd()
    if requirements_path is None:
        requirements_file = os.path.normpath(os.path.join(base_path, "./config.yml"))
    else:
        requirements_file = os.path.normpath(os.path.join(base_path, requirements_path))

    with open(requirements_file, "r") as f:
        requirements = yaml.safe_load(f)

    # Hash the requirements to avoid re-installing packages
    requirements_hash = str(hash(json.dumps(requirements)))
    if os.environ.get("requirements_hash") != requirements_hash:
        packages_default_common = requirements.get("default", {}).get("packages_common", []) or []
        packages_default_environment_specific = (
            requirements.get("default", {}).get(f"packages_{ENVIRONMENT.value}", []) or []
        )

        notebook_requirements = next(
            (cfg for cfg in requirements.get("notebooks", []) if cfg.get("name") == notebook_name), None
        )
        if notebook_requirements:
            packages_notebook_common = notebook_requirements.get("packages_common", []) or []
            packages_notebook_environment_specific = (
                notebook_requirements.get(f"packages_{ENVIRONMENT.value}", []) or []
            )
        else:
            packages_notebook_common = []
            packages_notebook_environment_specific = []

        # Note: environment specific packages have to be installed first,
        # because in Pyodide common packages might depend on them
        packages = [
            *packages_default_environment_specific,
            *packages_notebook_environment_specific,
            *packages_default_common,
            *packages_notebook_common,
        ]

        for pkg in packages:
            if ENVIRONMENT == EnvironmentEnum.PYODIDE:
                await install_package_pyodide(pkg, verbose)
            elif ENVIRONMENT == EnvironmentEnum.PYTHON:
                install_package_python(pkg, verbose)

        if verbose:
            print("Packages installed successfully.")
        os.environ["requirements_hash"] = requirements_hash


def set_data_pyodide(key: str, value: Any):
    """
    Take a Python object, serialize it to JSON, and send it to the host environment
    through a JavaScript function defined in the JupyterLite extension `data_bridge`.

    Args:
        key (string): The name under which data will be sent.
        value (Any): The value to send to the host environment.
    """
    serialized_data = json.dumps({key: value})
    js_code = f"""
      (function() {{
          if (window.sendDataToHost) {{
              window.sendDataToHost({serialized_data});
              console.log('Data sent to host:', {serialized_data});
          }} else {{
              console.error('sendDataToHost function is not defined on the window object.');
          }}
      }})();
      """
    display(Javascript(js_code))
    print(f"Status: {key} sent to host.")
    set_data_python(key, value)


def set_data_python(key: str, value: Any):
    """
    Write data to the `uploads` folder in a JupyterLab environment.
    Args:
        key (string): The name under which data will be written.
        value (Any): The value to write to the `uploads` folder.
    """
    if not os.path.exists(UPLOADS_FOLDER):
        os.makedirs(UPLOADS_FOLDER)
    for item in value:
        safe_name = item["name"].replace("%", "pct").replace("/", ":")
        file_path = os.path.join(UPLOADS_FOLDER, f"{safe_name}.json")
        with open(file_path, "w") as file:
            json.dump(item, file)
        print(f"Data for {key} written to {file_path}")


def set_data(key: str, value: Any):
    """
    Switch between the two functions `set_data_pyodide` and `set_data_python` based on the environment.
    Args:
        key (string): The name under which data will be written or sent.
        value (Any): The value to write or send.
    """
    if ENVIRONMENT == EnvironmentEnum.PYODIDE:
        set_data_pyodide(key, value)
    elif ENVIRONMENT == EnvironmentEnum.PYTHON:
        set_data_python(key, value)


def get_data_pyodide(key: str, globals_dict: Optional[Dict] = None):
    """
    Load data from the host environment into globals()[key] variable.
    Args:
        key (string): global variable name to store the received data.
        globals_dict (dict): globals() dictionary of the current scope.
    """
    if globals_dict is not None:
        globals_dict[key] = globals_dict["data_from_host"]


def get_data_python(key: str, globals_dict: Optional[Dict] = None):
    """
    Read data from the `uploads` folder in a JupyterLab environment.
    Args:
        key (string): The name under which data is expected to be received.
        globals_dict (dict): A dictionary to store the received data. Defaults to None.
    """
    try:
        data_from_host = []
        index = 0
        for filename in sorted(os.listdir(UPLOADS_FOLDER)):
            if filename.endswith(".json"):
                with open(os.path.join(UPLOADS_FOLDER, filename), "r") as file:
                    data = json.load(file)
                name = os.path.splitext(filename)[0]
                print(f"{index}: Data from {name} has been read successfully.")
                index += 1
                data_from_host.append(data)
        if globals_dict is not None:
            globals_dict[key] = data_from_host
    except FileNotFoundError:
        print("No data found in the 'uploads' folder.")


def get_data(key: str, globals_dict: Optional[Dict] = None):
    """
    Switch between the two functions `get_data_pyodide` and `get_data_python` based on the environment.
    Args:
        key (string): The name under which data is expected to be received.
        globals_dict (dict): A dictionary to store the received data. Defaults to None.
    """
    if ENVIRONMENT == EnvironmentEnum.PYODIDE:
        get_data_pyodide(key, globals_dict)
    elif ENVIRONMENT == EnvironmentEnum.PYTHON:
        get_data_python(key, globals_dict)



#!/usr/bin/env python
# coding: utf-8

# <a href="https://colab.research.google.com/github/Exabyte-io/api-examples/blob/dev/examples/job/run-simulations-and-extract-properties.ipynb" target="_parent">
# <img alt="Open in Google Colab" src="https://user-images.githubusercontent.com/20477508/128780728-491fea90-9b23-495f-a091-11681150db37.jpeg" width="150" border="0">
# </a>

# # Run Simulations and Extract Properties
#
# This example demonstrates how to use Mat3ra RESTful API to create simulation [Jobs](https://docs.mat3ra.com/jobs/overview/) programmatically for multiple [Materials](https://docs.mat3ra.com/materials/overview/) at once and extract the resulting [Properties](https://docs.mat3ra.com/properties/overview/) forming a [Pandas](https://pandas.pydata.org/) dataframe.
#
# This approach can work with any [Workflows](https://docs.mat3ra.com/workflows/overview/). For the demonstration purpose we use the Density Functional Theory and extract Electronic Band Gap as the property of interest.
#
# > <span style="color: orange">**IMPORTANT NOTE**</span>: In order to run this example in full, an active Mat3ra.com account with access to VASP (Vienna ab-initio simulations package) is required. Alternatively, Readers may substitute the workflow ID below with another one (an equivalent one for Quantum ESPRESSO, for example) and adjust extraction of the results ("Extract results" section). RESTful API credentials shall be updated in [settings](../../utils/settings.json).
#
#
# ## Steps
#
# We follow the below steps:
#
# - Import materials from [materials project](https://materialsproject.org/)
#
# - Group imported materials inside a [materials set](https://docs.mat3ra.com/entities-general/sets/)
#
# - Create jobs for the materials and grouping them inside a [jobs set](https://docs.mat3ra.com/entities-general/sets/)
#
# - Submit jobs and monitoring the progress
#
# - Extract the [final structure](https://docs.mat3ra.com/properties/structural/final-structure) (relaxed structure) and its properties
#
# - Output the results as Pandas dataFrame
#
# ## Pre-requisites
#
# The explanation below assumes that the reader is familiar with the concepts used in Mat3ra platform and RESTful API. We outline these below and direct the reader to the original sources of information:
#
# - [Generating RESTful API authentication parameters](../system/get_authentication_params.ipynb)
# - [Importing materials from materials project](../material/import_materials_from_materialsproject.ipynb)
# - [Creating and submitting jobs](../job/create_and_submit_job.ipynb)

# # Complete Authorization Form and Initialize Settings
#
# This will also determine environment and set all environment variables. We determine if we are using Jupyter Notebooks or Google Colab to run this tutorial.
#
# If you are running this notebook from Google Colab, Colab takes ~1 min to execute the following cell.
#
# ACCOUNT_ID and AUTH_TOKEN - Authentication parameters needed for when making requests to [Mat3ra.com's API Endpoints](https://docs.mat3ra.com/rest-api/endpoints/).
#
# MATERIALS_PROJECT_API_KEY - Authentication parameter needed for when making requests to [Material Project's API](https://materialsproject.org/open)
#
# ORGANIZATION_ID - Authentication parameter needed for when working with collaborative accounts https://docs.mat3ra.com/collaboration/organizations/overview/
#
# > <span style="color: orange">**NOTE**</span>: If you are running this notebook from Jupyter, the variables ACCOUNT_ID, AUTH_TOKEN, MATERIALS_PROJECT_API_KEY, and ORGANIZATION_ID should be set in the file [settings.json](../../utils/settings.json) if you need to use these variables. To obtain API token parameters, please see the following link to the documentation explaining how to get them: https://docs.mat3ra.com/accounts/ui/preferences/api/

# In[ ]:


# @title Authorization Form
ACCOUNT_ID = "ACCOUNT_ID"  # @param {type:"string"}
AUTH_TOKEN = "AUTH_TOKEN"  # @param {type:"string"}
MATERIALS_PROJECT_API_KEY = "MATERIALS_PROJECT_API_KEY"  # @param {type:"string"}
ORGANIZATION_ID = "ORGANIZATION_ID"  # @param {type:"string"}

import os

if "COLAB_JUPYTER_IP" in os.environ:
    os.environ.update(
        dict(
            ACCOUNT_ID=ACCOUNT_ID,
            AUTH_TOKEN=AUTH_TOKEN,
            MATERIALS_PROJECT_API_KEY=MATERIALS_PROJECT_API_KEY,
            ORGANIZATION_ID=ORGANIZATION_ID,
        )
    )

    get_ipython().system('GIT_BRANCH="dev"; export GIT_BRANCH; curl -s "https://raw.githubusercontent.com/Exabyte-io/api-examples/${GIT_BRANCH}/scripts/env.sh" | bash')


# ### Import packages

# In[ ]:


import time
from IPython.display import IFrame

# Import settings file and utils file
from utils.settings import ENDPOINT_ARGS, ACCOUNT_ID, MATERIALS_PROJECT_API_KEY
from utils.generic import wait_for_jobs_to_finish, get_property_by_subworkflow_and_unit_indicies, dataframe_to_html

import pandas as pd

# Relevant functions from the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.utils.materials import flatten_material
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints
from exabyte_api_client.endpoints.properties import PropertiesEndpoints


# #### Materials
#
# - **MATERIALS_PROJECT_IDS**: a list of material IDs to be imported from materials project
# - **TAGS**: a list of [tags](https://docs.mat3ra.com/entities-general/data/#tags) to assign to imported materials
# - **MATERIALS_SET_NAME**: the name of the materials set
#

# In[ ]:


MATERIALS_PROJECT_IDS = ["mp-149", "mp-32"]  # Si and Ge
MATERIALS_SET_NAME = "materials-set"
TAGS = ["tag1", "tag2"]


# #### Jobs
#
# Parameters for the jobs to be ran for the imported materials:
#
# - **JOB_NAME_PREFIX**: prefix to be used for the job name with "{JOB_NAME_PREFIX} {FORMULA}" convention (e.g.  "Job Name Prefix - SiGe")
# - **JOBS_SET_NAME**: the name of the jobs set

# In[ ]:


JOB_NAME_PREFIX = "Job Name Prefix"
JOBS_SET_NAME = "jobs-set"


# #### Workflow
#
# This example is based on [this](https://platform.mat3ra.com/bank/workflows/tPiF5dBQrY8pnik8r) bank workflow which is later copied to the account workflows collection.  The workflow is named "D3-GGA-BS-BG-DOS-ALL" and utilizes the logic explained in https://arxiv.org/pdf/1808.05325.pdf, for example (see section "Methodology", Table I). "D3" indicates the difficulty level 3 per the table convention. BS, BG, DOS indicate the properties extracted - Band Structure, Band Gap, Density of States. The workflow is utilizing VASP simulation engine at version 5.4.4.

#

# In[ ]:


BANK_WORKFLOW_ID = "tPiF5dBQrY8pnik8r"


# In[ ]:


# Visualize the bank workflow below
# NOTE: might not be rendered on Github
IFrame("https://platform.mat3ra.com/analytics/workflows/{}".format(BANK_WORKFLOW_ID), width=900, height=650)


# #### Compute
#
# Setup compute parameters. See [this](https://docs.mat3ra.com/infrastructure/compute-settings/ui) for more information about compute parameters.
#
# - **NODES**: Number of nodes. Defaults to 1.
# - **PPN**: Number of MPI processes per each node, Defaults to 1.
# - **QUEUE**: The name of queue to submit the jobs into. Defaults to D.
# - **TIME_LIMIT**: Job walltime. Defaults to "01:00:00" (one hour).
# - **CLUSTER**: The full qualified domain name (FQDN) or alias of the cluster to submit the jobs into.
#
# explain in the notebook that the job might run out of the limit of memory so it is clear, (2) suggest using OR queue to avoid memory limitations
#
# > <span style="color: orange">**NOTE**</span>: Although here we set the QUEUE to be debug, it is possible the job might run out of memory, and result in an Errored-Jobs status. If this happens, we suggest you switch from `QUEUE = D` to `QUEUE = OR` to avoid memory limitations.

# In[ ]:


PPN = "1"
QUEUE = "D"
NODES = "1"
TIME_LIMIT = "01:00:00"
CLUSTER = "cluster-001"


# ### Initialize endpoints

# In[ ]:


job_endpoints = JobEndpoints(*ENDPOINT_ARGS)
project_endpoints = ProjectEndpoints(*ENDPOINT_ARGS)
material_endpoints = MaterialEndpoints(*ENDPOINT_ARGS)
property_endpoints = PropertiesEndpoints(*ENDPOINT_ARGS)
bank_workflow_endpoints = BankWorkflowEndpoints(*ENDPOINT_ARGS)


# Next, we retrieve the owner and project IDs as they are needed by the endpoints. Account's default material is used to extract the owner ID. One can extract the owner ID from any other account's [entities](https://docs.mat3ra.com/entities-general/overview/).

# In[ ]:


owner_id = material_endpoints.list({"isDefault": True, "owner._id": ACCOUNT_ID})[0]["owner"]["_id"]
project_id = project_endpoints.list({"isDefault": True, "owner._id": ACCOUNT_ID})[0]["_id"]


# ### Create workflow
#
# Copy bank workflow (template) to the account's workflows collection.

# In[ ]:


workflow_id = bank_workflow_endpoints.copy(BANK_WORKFLOW_ID, owner_id)["_id"]


# ### Import materials
#
# Import materials from materials project with the above tags.

# In[ ]:


materials = material_endpoints.import_from_materialsproject(
    MATERIALS_PROJECT_API_KEY, MATERIALS_PROJECT_IDS, owner_id, TAGS
)


# Create a materials set and move the materials into it.

# In[ ]:


materials_set = material_endpoints.create_set({"name": MATERIALS_SET_NAME, "owner": {"_id": owner_id}})
for material in materials:
    material_endpoints.move_to_set(material["_id"], "", materials_set["_id"])


# ### Create jobs
#
# Create jobs for the materials above.

# In[ ]:


compute = job_endpoints.get_compute(CLUSTER, PPN, NODES, QUEUE, TIME_LIMIT)
jobs = job_endpoints.create_by_ids(materials, workflow_id, project_id, owner_id, JOB_NAME_PREFIX, compute)


# Create a jobs set and move the jobs into it.

# In[ ]:


jobs_set = job_endpoints.create_set({"name": JOBS_SET_NAME, "projectId": project_id, "owner": {"_id": owner_id}})
for job in jobs:
    job_endpoints.move_to_set(job["_id"], "", jobs_set["_id"])


# Submit the jobs for execution.

# In[ ]:


for job in jobs:
    job_endpoints.submit(job["_id"])


# Monitor the jobs and print the status until they are all finished.

# In[ ]:


job_ids = [job["_id"] for job in jobs]
wait_for_jobs_to_finish(job_endpoints, job_ids)


# ### Extract results
#
# For each material, simulaion job, final structure, pressure and band gaps are extracted.
#
# - Final structure and pressure are extracted from the first unit (vasp_relax with index 0) of the first job's subworkflow (volume-relaxation with index 0)
#
# - Band gaps are extracted from the second unit (vasp-bands with index 1) of the second job's subworkflow (SCF-BS-BG-DOS with index 1).

# In[ ]:


results = []
for material in materials:
    job = next((job for job in jobs if job["_material"]["_id"] == material["_id"]))
    final_structure = get_property_by_subworkflow_and_unit_indicies(
        property_endpoints, "final_structure", job, 0, 0
    )["data"]
    pressure = get_property_by_subworkflow_and_unit_indicies(property_endpoints, "pressure", job, 0, 0)["data"][
        "value"
    ]
    unit_flowchart_id = job["workflow"]["subworkflows"][1]["units"][1]["flowchartId"]
    band_gap_direct = property_endpoints.get_direct_band_gap(job["_id"], unit_flowchart_id)
    band_gap_indirect = property_endpoints.get_indirect_band_gap(job["_id"], unit_flowchart_id)
    results.append(
        {
            "initial_structure": material,
            "final_structure": final_structure,
            "pressure": pressure,
            "band_gap_direct": band_gap_direct,
            "band_gap_indirect": band_gap_indirect,
        }
    )


# ### Flatten results
#
# The below for-loop iterates over the results and flatten them to form the final Pandas dataFrame.

# In[ ]:


table = []
for result in results:
    data = flatten_material(result["initial_structure"])
    data.extend(flatten_material(result["initial_structure"]))
    data.extend([result["pressure"], result["band_gap_direct"], result["band_gap_indirect"]])
    table.append(data)


# ### Output results
#
# Form the Pandas dataFrame headers according to the table generated above with the following abbreviations:
#
# - **"INI"**: INITIAL
# - **"FIN"**: FINAL
# - **"N-SITES"**: Number of Sites
# - **"LAT"**: LATTICE

# In[ ]:


headers = []
keys = ["ID", "NAME", "TAGS", "NS", "LAT-A", "LAT-B", "LAT-C", "LAT-ALPHA", "LAT-BETA", "LAT-GAMMA"]
headers.extend(["-".join(("INI", key)) for key in keys])
headers.extend(["-".join(("FIN", key)) for key in keys])
headers.extend(["PRESSURE", "DIRECT-GAP", "INDIRECT-GAP"])


# Create and print the final table as Pandas dataFrame.

# In[ ]:


df = pd.DataFrame(data=table, columns=headers)
html = dataframe_to_html(df)
html




#!/usr/bin/env python
# coding: utf-8

# # Overview
# 
# This example has been created as part of our Advanced Topics Webinar, scheduled for March 26, 2021. This webinar focused on the high-throughput screening of surfaces using Exabyte.
# 
# In this notebook, a structure for γ-Al2O3, an industrially-relevant material used in a variety of catalytic applications, both as a catalyst and as a a support. One such application is the [Claus Process](https://en.wikipedia.org/wiki/Claus_process), which is catalyzed by γ-Al2O3 and produced over 64-million tons of sulfur in 2005!

# # Execution
# 
# > <span style="color: orange">**NOTE**</span>: In order to run this example, an active Exabyte.io account is required. RESTful API credentials shall be updated in [settings](../settings.py). The generation of the credentials is also explained therein.
# 
# In addition, this notebook demonstrates the use of an "Organization ID," which allows multiple users who are part of the same organization to collaborate.
# 
# ## Import packages

# In[]:


import urllib
import os, sys

import ase.io
import ase.build.tools
import ase.cluster
from ase.visualize import view

import pymatgen.symmetry.analyzer

# Import settings file and utils
module_path = os.path.abspath(os.path.join('..'))
if module_path not in sys.path:
    sys.path.append(module_path)
from utils.generic import ensure_packages_are_installed, display_JSON, save_files
ensure_packages_are_installed()
from utils.material import get_all_slabs_and_terms, freeze_center_bulk
from settings import ENDPOINT_ARGS, ORGANIZATION_ID

# Import relevant portions of the API client
from exabyte_api_client.endpoints.jobs import JobEndpoints
from exabyte_api_client.endpoints.projects import ProjectEndpoints
from exabyte_api_client.endpoints.materials import MaterialEndpoints
from exabyte_api_client.endpoints.workflows import WorkflowEndpoints
from exabyte_api_client.endpoints.bank_workflows import BankWorkflowEndpoints


# # Visualize the Unit Cell
# 
# We begin by finding the following unit cell:
# - Gamma Al2O3, whose unit cell is in a POSCAR we constructed based on Digne, M.; Sautet, P.; Raybaud, P.; Euzen, P.; Toulhoat, H. Use of DFT to achieve a rational understanding of acid-basic properties of γ-alumina surfaces. J Catal 2004, 226, 54-68.
# 
# For this example, we have already taken the unit cell information from the literature and created a POSCAR file from it.

# In[]:


# Get gamma Al2O3 from the local disk
al2o3_poscar_filename = "../assets/gamma_alumina_digne_et_al.poscar"
al2o3_ase = ase.io.read(al2o3_poscar_filename)

# View the crystal in ASE's built-in x3d viewer
# The viewer is interactive. Press "I" to center the cell followed by "A" to zoom out
view(al2o3_ase, viewer='x3d')


# # Optimize the Unit Cells

# The next few steps will focus on optimizing the unit cell with our chosen DFT methodology, before we cleave it.
# 
# ## Create the Material
# Now that we have a unit cell, we can upload it to the platform.

# In[]:


# Create our Materials Endpoint
exabyte_materials_endpoint = MaterialEndpoints(*ENDPOINT_ARGS)

#Upload Al2O3 to the platform
with open(al2o3_poscar_filename, "r") as inp:
    content = inp.read()
    al2o3_cell_material = exabyte_materials_endpoint.import_from_file(name="Gamma_Alumina_Unit_Cell",
                                                                      content=content,
                                                                      owner_id = ORGANIZATION_ID)


# ## Create the Workflow
# Create a workflow based on the Variable-cell Relaxation workflow available [here](https://platform.exabyte.io/analytics/workflows/NAdKjws8qieKWeYnL)

# In[]:


# Start by finding a unit cell optimization workflow on the Exabyte platform.
# The bank workflow ID can be found in the bank workflows URL
exabyte_bank_workflows_endpoint = BankWorkflowEndpoints(*ENDPOINT_ARGS)
bank_workflow_id = "NAdKjws8qieKWeYnL"
al2o3_workflow = exabyte_bank_workflows_endpoint.copy(bank_workflow_id, account_id = ORGANIZATION_ID)

# Set the names to something easy to recognize, and upload
exabyte_workflows_endpoint = WorkflowEndpoints(*ENDPOINT_ARGS)
al2o3_workflow['name'] = "Alumina_Cell_Relax"
al2o3_cellopt_workflow = exabyte_workflows_endpoint.create(al2o3_workflow)


# ## Create the Job
# Then create the jobs to optimize these unit cells

# In[]:


# Get the default projectID for the organization account (this can be substituted with a userID)
exabyte_projects_endpoint = ProjectEndpoints(*ENDPOINT_ARGS)
project_id = exabyte_projects_endpoint.list({"isDefault": True,
                                             "owner._id": ORGANIZATION_ID})[0]["_id"]

# Create the compute configuration for the jobs on Azure
exabyte_jobs_endpoint = JobEndpoints(*ENDPOINT_ARGS)
job_config = {"ppn": 16,
              "queue": "OF",
              "nodes": 1,
              "time_limit": "00:20:00",
              "cluster": "cluster-007"}
compute = exabyte_jobs_endpoint.get_compute(**job_config)

# Create the Al2O3 job
al2o3_job = exabyte_jobs_endpoint.create_by_ids([al2o3_cell_material],
                                                al2o3_cellopt_workflow["_id"],
                                                project_id,
                                                ORGANIZATION_ID,
                                                "al2o3_cellopt",
                                                compute)


# Finally, submit the jobs

# In[]:


exabyte_jobs_endpoint.submit(al2o3_job[0]["_id"])


# # Generate the Slabs

# We begin by extracting the relaxed unit cell from the jobs

# In[]:


# Extract job ID
al2o3_job_id = al2o3_job[0]["_id"]

save_files(al2o3_job_id, exabyte_jobs_endpoint, "CONTCAR", "al2o3_relaxed.vasp")


# The below code uses Pymatgen to find all unique planes in a crystal. It then generates slabs with every possible termination in that plane. Finally, asymmetric slabs are filtered out, and the slabs are saved to a dictionary. The dictionary's format is: `{miller-index: {termination: {"slab": slab} }`. Note that here, to take advantage of several convenient functions in ASE, we have the function output ASE Atoms objects instead of PyMatGen objects.

# In[]:


al2o3_relaxed_pymatgen = pymatgen.core.structure.Structure.from_file("al2o3_relaxed.vasp")

al2o3_slabs = get_all_slabs_and_terms(al2o3_relaxed_pymatgen, thickness=3, is_by_layers=True)


# # Write the Slabs to Disk
# 
# To finish preparing our structures for VASP, we do the following for both sets of slabs:
# 1. Center the slab, and adjust the vacuum to 10Å
# 2. Freeze the center layer of the slab to simulate the bulk
# 3. Write the slab to a vasp-format file

# In[]:


# Declare the vacuum size we want to have (for easy adjustment by future users)
vacuum_size = 10
for miller_index, term_dict in al2o3_slabs.items():
    for term, surface in term_dict.items():
        # Sort the slab by symbol, to help make the POTCAR file smaller
        slab = ase.build.tools.sort(surface["slab"])

        # Center the slab's coordinates and adjust the vacuum to vacuum_size
        # Note that the "vacuum" argument refers to the amount of vacuum on either side
        # So, we divide by 2
        slab.center(vacuum=vacuum_size/2, axis=2)

        # Add a FixAtoms constraint to freeze the center layer of atoms
        freeze_center_bulk(slab)

        # Write the slab to a file
        formula = slab.get_chemical_formula(empirical=True)
        filename = f"{formula}_{miller_index}_term{term}.vasp"
        ase.io.write(filename, slab)


# # Upload POSCARS
# 
# Now that we have a set of prepared POSCARs, we can upload them to our user account on Exabyte. Let's create a couple material sets to hold these.

# In[]:


al2o3_materials_set = exabyte_materials_endpoint.create_set({"name" : "Webinar_Slabs_Al2O3",
                                                             "owner" : {"_id" : ORGANIZATION_ID}})


# And finally, we will upload all of the slabs to the platform

# In[ ]:


for miller_index, term_dict in al2o3_slabs.items():
    for term, surface in term_dict.items():
        slab = surface["slab"]

        # Get the filename we generated above
        formula = slab.get_chemical_formula(empirical=True)
        material_name = f"{formula}_{miller_index}_term{term}"
        filename = f"{material_name}.vasp"

        # Import the material to Exabyte, and place it in the correct material set
        with open(filename, "r") as inp:
            content = inp.read()
        material_json = exabyte_materials_endpoint.import_from_file(name=material_name, content=content,
                                                                    owner_id=ORGANIZATION_ID)
        exabyte_materials_endpoint.move_to_set(material_json["_id"], "", al2o3_materials_set["_id"])


# In[ ]:




